import os
import warnings
from datetime import datetime
from unittest import skipIf

import pytest
import pytz
import requests

from pykechain.enums import ActivityType, ActivityStatus
from pykechain.exceptions import NotFoundError, MultipleFoundError, IllegalArgumentError
from pykechain.models import Activity
from pykechain.utils import temp_chdir
from tests.classes import TestBetamax
from tests.utils import TEST_FLAG_IS_WIM2

ISOFORMAT = "%Y-%m-%dT%H:%M:%SZ"
ISOFORMAT_HIGHPRECISION = "%Y-%m-%dT%H:%M:%S.%fZ"


class TestActivities(TestBetamax):
    def test_retrieve_activities(self):
        self.assertTrue(self.project.activities())

    def test_retrieve_single_activity(self):
        self.assertTrue(self.project.activity('Specify wheel diameter'))

    def test_activity_attributes(self):
        attributes = ['_client', '_json_data', 'id', 'name', 'created_at', 'updated_at', 'ref',
                      'description', 'status', 'activity_type', '_scope_id',
                      'start_date', 'due_date']

        obj = self.project.activity('Specify wheel diameter')
        for attribute in attributes:
            self.assertTrue(hasattr(obj, attribute),
                            "Could not find '{}' in the object: '{}'".format(attribute, obj.__dict__.keys()))

    def test_retrieve_unknown_activity(self):
        with self.assertRaises(NotFoundError):
            self.project.activity('Hello?!')

    def test_retrieve_too_many_activity(self):
        with self.assertRaises(MultipleFoundError):
            self.project.activity()

    def test_create_activity(self):
        # set up
        project = self.project

        subprocess = project.create_activity('Test subprocess creation', activity_type='PROCESS')
        self.assertEqual(subprocess.name, 'Test subprocess creation')

        task = subprocess.create('Test task creation')
        self.assertEqual(task.name, 'Test task creation')

        # teardown
        subprocess.delete()

        with self.assertRaises(NotFoundError):
            project.activity(name='Test subprocess creation')
        with self.assertRaises(NotFoundError):
            project.activity(name='Test task creation')

    def test_create_activity_under_task(self):
        task = self.project.activity('Specify wheel diameter')

        with self.assertRaises(IllegalArgumentError):
            task.create('This cannot happen')

    # new in 1.7
    def test_edit_activity_name(self):
        specify_wd = self.project.activity('Specify wheel diameter')
        specify_wd.edit(name='Specify wheel diameter - updated')

        specify_wd_u = self.project.activity('Specify wheel diameter - updated')
        self.assertEqual(specify_wd.id, specify_wd_u.id)
        self.assertEqual(specify_wd.name, specify_wd_u.name)
        self.assertEqual(specify_wd.name, 'Specify wheel diameter - updated')

        # Added to improve coverage. Assert whether IllegalArgumentError is raised when 'name' is not a string object.
        with self.assertRaises(IllegalArgumentError):
            specify_wd.edit(name=True)

        specify_wd.edit(name='Specify wheel diameter')

    def test_edit_activity_description(self):
        specify_wd = self.project.activity('Specify wheel diameter')
        specify_wd.edit(description='This task has an even cooler description')

        self.assertEqual(specify_wd._client.last_response.status_code, requests.codes.ok)

        # Added to improve coverage. Assert whether IllegalArgumentError is raised when 'description' is
        # not a string object.
        with self.assertRaises(IllegalArgumentError):
            specify_wd.edit(description=42)

        specify_wd.edit(description='This task has a cool description')

    def test_edit_activity_naive_dates(self):
        specify_wd = self.project.activity('Specify wheel diameter')

        old_start, old_due = self._convert_timestamp(specify_wd._json_data.get('start_date')), \
                             self._convert_timestamp(specify_wd._json_data.get('due_date'))
        start_time = datetime(2000, 1, 1, 0, 0, 0)
        due_time = datetime(2019, 12, 31, 0, 0, 0)

        with warnings.catch_warnings(record=False) as w:
            warnings.simplefilter("ignore")
            specify_wd.edit(start_date=start_time, due_date=due_time)

        self.assertEqual(specify_wd._client.last_response.status_code, requests.codes.ok)

        # Added to improve coverage. Assert whether IllegalArgumentError is raised when 'start_date' and
        # 'due_date are not datetime objects
        with self.assertRaises(IllegalArgumentError):
            specify_wd.edit(start_date='All you need is love')

        with self.assertRaises(IllegalArgumentError):
            specify_wd.edit(due_date='Love is all you need')

        specify_wd_u = self.project.activity('Specify wheel diameter')

        self.assertEqual(specify_wd.id, specify_wd_u.id)

        specify_wd.edit(start_date=old_start, due_date=old_due)

    def test_edit_due_date_timezone_aware(self):
        specify_wd = self.project.activity('Specify wheel diameter')

        # save old values
        old_start, old_due = self._convert_timestamp(specify_wd._json_data.get('start_date')), \
                             self._convert_timestamp(specify_wd._json_data.get('due_date'))

        startdate = datetime.now(pytz.utc)
        duedate = datetime(2019, 1, 1, 0, 0, 0, tzinfo=pytz.timezone('Europe/Amsterdam'))

        specify_wd.edit(start_date=startdate, due_date=duedate)

        self.assertEqual(specify_wd._client.last_response.status_code, requests.codes.ok)

        # teardown
        specify_wd.edit(start_date=old_start, due_date=old_due)

    # 1.10.0
    def test_edit_activity_status(self):
        specify_wd = self.project.activity('Specify wheel diameter')
        original_status = specify_wd.status

        specify_wd.edit(status=ActivityStatus.COMPLETED)

        # Added to improve coverage. Assert whether IllegalArgumentError is raised when 'status' is not a string
        with self.assertRaises(IllegalArgumentError):
            specify_wd.edit(status=True)

        # If the status is not part of Enums.Status then it should raise an APIError
        with self.assertRaises(IllegalArgumentError):
            specify_wd.edit(status='NO STATUS')

        # Return the status to how it used to be
        specify_wd.edit(status=original_status)

    def _convert_timestamp(self, value):
        try:
            r = datetime.strptime(value, ISOFORMAT)
        except ValueError:
            r = datetime.strptime(value, ISOFORMAT_HIGHPRECISION)
        if isinstance(r, datetime):
            return r
        else:
            raise ValueError

    # 1.7.2
    def test_datetime_with_naive_duedate_only_fails(self):
        """reference to #121 - thanks to @joost.schut"""
        # setup
        specify_wd = self.project.activity('Specify wheel diameter')

        # save old values
        old_start, old_due = self._convert_timestamp(specify_wd._json_data.get('start_date')), \
                             self._convert_timestamp(specify_wd._json_data.get('due_date'))
        naive_duedate = datetime(2017, 6, 5, 5, 0, 0)
        with warnings.catch_warnings(record=False) as w:
            warnings.simplefilter("ignore")
            specify_wd.edit(due_date=naive_duedate)

        # teardown
        with warnings.catch_warnings(record=False) as w:
            warnings.simplefilter("ignore")
            specify_wd.edit(due_date=old_due)

    def test_datetime_with_tzinfo_provides_correct_offset(self):
        """reference to #121 - thanks to @joost.schut

        The tzinfo.timezone('Europe/Amsterdam') should provide a 2 hour offset, recording 20 minutes
        """
        # setup
        specify_wd = self.project.activity('Specify wheel diameter')
        # save old values
        old_start, old_due = self._convert_timestamp(specify_wd._json_data.get('start_date')), \
                             self._convert_timestamp(specify_wd._json_data.get('due_date'))

        tz = pytz.timezone('Europe/Amsterdam')
        tzaware_due = tz.localize(datetime(2017, 7, 1))
        tzaware_start = tz.localize(datetime(2017, 6, 30, 0, 0, 0))

        specify_wd.edit(start_date=tzaware_start)
        self.assertTrue(specify_wd._json_data['start_date'], tzaware_start.isoformat(sep='T'))

        specify_wd.edit(due_date=tzaware_due)
        self.assertTrue(specify_wd._json_data['due_date'], tzaware_due.isoformat(sep='T'))

        # teardown
        with warnings.catch_warnings(record=False) as w:
            warnings.simplefilter("ignore")
            specify_wd.edit(start_date=old_start, due_date=old_due)

    def test_retrieve_children_of_task_fails_for_task(self):
        task = self.project.activity(name='Specify wheel diameter')
        with self.assertRaises(NotFoundError):
            task.children()

    def test_retrieve_activity_by_id(self):
        task = self.project.activity(name='Subprocess')  # type: Activity

        task_by_id = self.client.activity(pk=task.id)

        self.assertEqual(task.id, task_by_id.id)

    def test_retrieve_siblings_of_a_task_in_a_subprocess(self):
        task = self.project.activity(name='Subprocess')  # type: Activity
        siblings = task.siblings()

        self.assertIn(task.id, [sibling.id for sibling in siblings])
        self.assertTrue(len(siblings) >= 1)

    # in 1.12

    def test_retrieve_siblings_of_a_task_in_a_subprocess_with_arguments(self):
        task = self.project.activity(name='SubTask')  # type: Activity
        siblings = task.siblings(name__icontains='sub')

        self.assertIn(task.id, [sibling.id for sibling in siblings])
        self.assertEqual(1, len(siblings))

    @skipIf(not TEST_FLAG_IS_WIM2, reason="This tests is designed for WIM version 2, expected to fail on old WIM")
    def test_activity2_without_scope_id_will_fix_itself(self):
        specify_wheel_diam_cripled = self.project.activity(name='Specify wheel diameter', fields='id,name,status')
        self.assertFalse(specify_wheel_diam_cripled._json_data.get('scope_id'))

        # now the self-healing will beging
        self.assertEqual(specify_wheel_diam_cripled.scope_id, self.project.id)

    # in 1.13
    def test_create_activity_with_incorrect_activity_class_fails(self):
        with self.assertRaisesRegex(IllegalArgumentError, 'Please provide accepted activity_type'):
            self.project.create_activity(name='New', activity_type='DEFUNCTActivity')


@skipIf(not TEST_FLAG_IS_WIM2, reason="This tests is designed for WIM version 2, expected to fail on older WIM")
class TestActivity2SpecificTests(TestBetamax):
    # 2.0 new activity
    def test_edit_activity2_assignee(self):
        specify_wd = self.project.activity('Specify wheel diameter')  # type: Activity2
        original_assignee_ids = specify_wd._json_data.get('assignee_ids') or []

        # pykechain_user = self.client.user(username='pykechain')
        test_user = self.client.user(username='testuser')

        specify_wd.edit(assignees_ids=[test_user.id])
        specify_wd.refresh()

        self.assertIsInstance(specify_wd._json_data.get('assignees_ids')[0], int)

        self.assertEqual(specify_wd._client.last_response.status_code, requests.codes.ok)

        # Added to improve coverage. Assert whether NotFoundError is raised when 'assignee' is not part of the
        # scope members
        with self.assertRaises(NotFoundError):
            specify_wd.edit(assignees_ids=[-100])

        # Added to improve coverage. Assert whether NotFoundError is raised when 'assignee' is not part of the
        # scope members
        with self.assertRaises(IllegalArgumentError):
            specify_wd.edit(assignees_ids='this should have been a list')

        specify_wd.edit(assignees_ids=original_assignee_ids)

    def test_activity2_retrieve_parent_of_task(self):
        task = self.project.activity(name='SubTask')
        subprocess = task.parent()  # type Activity
        self.assertEqual(subprocess.activity_type, ActivityType.PROCESS)

    def test_activity2_retrieve_parent_of_a_toplevel_task_returns_workflow_root_id(self):
        task = self.project.activity('Specify wheel diameter')
        parent = task.parent()
        self.assertEqual(self.project._json_data.get('workflow_root_id'), parent.id)

    def test_activity2_test_workflow_root_object(self):
        workflow_root = self.project.activity(id=self.project._json_data.get('workflow_root_id'))

        self.assertTrue(workflow_root.is_root())
        self.assertTrue(workflow_root.is_workflow_root())

    def test_activity2_retrieve_children_of_parent(self):
        subprocess = self.project.activity(name='Subprocess')  # type: Activity
        children = subprocess.children()
        self.assertTrue(len(children) >= 1)
        for child in children:
            self.assertEqual(child._json_data.get('parent_id'), subprocess.id)

    def test_activity2_retrieve_children_of_subprocess_with_arguments(self):
        subprocess = self.project.activity(name='Subprocess')  # type: Activity
        children = subprocess.children(name__icontains='task')
        self.assertTrue(len(children) >= 1)
        for child in children:
            self.assertEqual(child._json_data.get('parent_id'), subprocess.id)

    def test_rootlevel_activity2_is_rootlevel(self):
        specify_wd = self.project.activity('Specify wheel diameter')

        self.assertTrue(specify_wd.is_rootlevel())

    def test_subtask_activity2_is_not_rootlevel(self):
        subprocess_subtask = self.project.activity('SubTask')

        self.assertFalse(subprocess_subtask.is_rootlevel())
        # self.assertTrue(subprocess_subtask.subprocess())

    def test_activity2_is_task(self):
        specify_wd = self.project.activity('Specify wheel diameter')

        self.assertTrue(specify_wd.is_task())
        self.assertFalse(specify_wd.is_subprocess())

    def test_activity2_is_subprocess(self):
        subprocess = self.project.activity('Subprocess')

        self.assertTrue(subprocess.is_subprocess())
        self.assertFalse(subprocess.is_task())

    def test_activity2_assignees_list(self):
        activity_name = 'test task'
        activity = self.project.activity(name=activity_name)  # type: Activity2

        list_of_assignees_in_data = activity._json_data.get('assignees_ids')
        assignees_list = activity.assignees

        self.assertSetEqual(set(list_of_assignees_in_data), set([u.id for u in assignees_list]))

    def test_activity2_assignees_list_no_assignees_gives_empty_list(self):
        activity_name = 'Specify wheel diameter'
        activity = self.project.activity(name=activity_name)  # type: Activity2

        list_of_assignees_in_data = activity._json_data.get('assignees_ids')
        assignees_list = activity.assignees

        self.assertListEqual(list(), activity.assignees, "Task has no assingees and should return Empty list")

    def test_activity2_move(self):
        # setUp
        activity_name = 'test task'
        activity_to_be_moved = self.project.activity(name=activity_name)
        original_parent = activity_to_be_moved.parent()

        new_parent_name = 'Subprocess'
        new_parent = self.project.activity(name=new_parent_name)

        activity_to_be_moved.move(parent=new_parent)

        activity_to_be_moved.refresh()

        # testing
        self.assertTrue(activity_to_be_moved.parent().id == new_parent.id)

        # tearDown
        activity_to_be_moved.move(parent=original_parent.id)
        activity_to_be_moved.refresh()
        self.assertTrue(activity_to_be_moved.parent().id == original_parent.id)

    def test_activity2_move_under_task_parent(self):
        # setUp
        activity_name = 'test task'
        activity_to_be_moved = self.project.activity(name=activity_name)

        new_parent_name = 'Specify wheel diameter'
        new_parent = self.project.activity(name=new_parent_name)

        # testing
        with self.assertRaises(IllegalArgumentError):
            activity_to_be_moved.move(parent=new_parent)

    def test_activity2_move_under_part_object(self):
        # setUp
        activity_name = 'test task'
        activity_to_be_moved = self.project.activity(name=activity_name)

        new_parent_name = 'Bike'
        new_parent = self.project.part(name=new_parent_name)

        # testing
        with self.assertRaises(IllegalArgumentError):
            activity_to_be_moved.move(parent=new_parent)


# @skip('Does not work in PIM2 until KEC-19193 is resolved')
class TestActivityDownloadAsPDF(TestBetamax):

    def test_activity2_download_as_pdf(self):
        # setUp
        activity_name = 'Task - Form'
        activity = self.project.activity(name=activity_name)

        # testing
        with temp_chdir() as target_dir:
            activity.download_as_pdf(target_dir=target_dir, pdf_filename='pdf_file')
            activity.download_as_pdf(target_dir=target_dir)
            pdf_file = os.path.join(target_dir, 'pdf_file.pdf')
            pdf_file_called_after_activity = os.path.join(target_dir, activity_name + '.pdf')
            self.assertTrue(pdf_file)
            self.assertTrue(pdf_file_called_after_activity)

    @pytest.mark.skipif("os.getenv('TRAVIS', False)",
                        reason="Skipping tests when using Travis, as Async PDF is not doable without being live")
    def test_activity2_download_as_pdf_async(self):
        activity_name = 'Task - Form + Tables + Service'
        activity = self.project.activity(name=activity_name)

        # testing
        with temp_chdir() as target_dir:
            activity.download_as_pdf(target_dir=target_dir, pdf_filename='pdf_file', include_appendices=True)
            pdf_file = os.path.join(target_dir, 'pdf_file.pdf')
            pdf_file_called_after_activity = os.path.join(target_dir, activity_name + '.pdf')
            self.assertTrue(pdf_file)
            self.assertTrue(pdf_file_called_after_activity)





from jsonschema import validate, ValidationError

from pykechain.enums import PropertyVTypes, PropertyType
from pykechain.exceptions import IllegalArgumentError
from pykechain.models import Property2, AttachmentProperty2
from pykechain.models.validators import PropertyValidator, ValidatorEffect, VisualEffect, ValidVisualEffect, \
    InvalidVisualEffect, NumericRangeValidator, BooleanFieldValidator
from pykechain.models.validators.validator_schemas import options_json_schema, validator_jsonschema_stub, \
    effects_jsonschema_stub
from pykechain.models.validators.validators import RegexStringValidator, RequiredFieldValidator, EvenNumberValidator, \
    OddNumberValidator, SingleReferenceValidator, EmailValidator, AlwaysAllowValidator, FileSizeValidator, \
    FileExtensionValidator
from tests.classes import SixTestCase, TestBetamax
from jsonschema import validate, ValidationError

from pykechain.enums import PropertyVTypes, PropertyType
from pykechain.exceptions import IllegalArgumentError
from pykechain.models import Property2, AttachmentProperty2
from pykechain.models.validators import PropertyValidator, ValidatorEffect, VisualEffect, ValidVisualEffect, \
    InvalidVisualEffect, NumericRangeValidator, BooleanFieldValidator
from pykechain.models.validators.validator_schemas import options_json_schema, validator_jsonschema_stub, \
    effects_jsonschema_stub
from pykechain.models.validators.validators import RegexStringValidator, RequiredFieldValidator, EvenNumberValidator, \
    OddNumberValidator, SingleReferenceValidator, EmailValidator, AlwaysAllowValidator, FileSizeValidator, \
    FileExtensionValidator
from tests.classes import SixTestCase, TestBetamax


class TestValidatorJSON(SixTestCase):
    def test_valid_numeric_range_validator_json(self):
        options = dict(
            validators=[dict(
                vtype='numericRangeValidator',
                config=dict(
                    minvalue=2,
                    maxvalue=10,
                    stepsize=2,
                    on_valid=[dict(effect="CssEffect", config=dict(applyCss="valid"))],
                    on_invalid=[dict(effect="CssEffect", config=dict(applyCss="invalid")),
                                dict(effect="ErrorText",
                                     config=dict(text="Range should be between 2 and 10 with step 2."))]
                )
            )]
        )

        validate(options_json_schema, options)

    def test_valid_requiredfield_validator_json(self):
        options = {'validators': [
            {
                'vtype': 'requiredFieldValidator',
                'config': {
                    'effects_when_valid': [],
                    'effects_when_invalid': [{'effect': "ErrorText", 'config': {'Text': "This field is required."}}]
                }
            }, {
                'vtype': 'numericRangeValidator',
                'config': {'minValue': 0}
            }
        ]}

        validate(options_json_schema, options)

    def test_valid_booleanfield_validator_json(self):
        options = {'validators': [{
            'vtype': 'booleanFieldValidator',
            'config': {
                'is_valid': False,
                'is_invalid': [True, None],
                'effects_when_valid': [],
                'effects_when_invalid': [{'effect': 'ErrorText', 'text': 'Value of the field should be False'}]}
        }]}

        validate(options_json_schema, options)

    def test_validator_invalid_vtype(self):
        validator_json = {
            'vtype': 'invalid',
            'config': {}
        }
        with self.assertRaisesRegex(ValidationError, "'invalid' is not one of"):
            validate(validator_json, validator_jsonschema_stub)

    def test_validator_missing_vtype(self):
        validator_json = {
            'config': {}
        }
        with self.assertRaisesRegex(ValidationError, "'vtype' is a required property"):
            validate(validator_json, validator_jsonschema_stub)

    def test_validator_config_not_a_dict(self):
        v = {
            'vtype': PropertyVTypes.NUMERICRANGE,
            'config': []
        }
        with self.assertRaisesRegex(ValidationError, "is not of type 'object'"):
            validate(v, validator_jsonschema_stub)

    def test_validator_on_valid_not_a_list(self):
        v = {
            'vtype': PropertyVTypes.NUMERICRANGE,
            'config': {
                'on_valid': {}
            }
        }
        with self.assertRaisesRegex(ValidationError, "is not of type 'array'"):
            validate(v, validator_jsonschema_stub)

    def test_validator_on_invalid_not_a_list(self):
        v = {
            'vtype': PropertyVTypes.NUMERICRANGE,
            'config': {
                'on_invalid': {}
            }
        }
        with self.assertRaisesRegex(ValidationError, "is not of type 'array'"):
            validate(v, validator_jsonschema_stub)

    def test_validator_on_valid_list_with_obj(self):
        v = {
            'vtype': PropertyVTypes.NUMERICRANGE,
            'config': {
                'on_valid': [{'effect': '', 'config': {}}]
            }
        }
        validate(v, validator_jsonschema_stub)

    def test_validator_on_invalid_list_with_obj(self):
        v = {
            'vtype': PropertyVTypes.NUMERICRANGE,
            'config': {
                'on_invalid': [{'effect': '', 'config': {}}]
            }
        }
        validate(v, validator_jsonschema_stub)

    def test_validatoreffect_requires_effect_property(self):
        v = {}
        with self.assertRaisesRegex(ValidationError, "'effect' is a required property"):
            validate(v, effects_jsonschema_stub)

    def test_validatoreffect_requires_config_property(self):
        v = {'effect': ''}
        with self.assertRaisesRegex(ValidationError, "'config' is a required property"):
            validate(v, effects_jsonschema_stub)

    def test_validatoreffect_not_allows_additional_properties(self):
        v = {
            'effect': '',
            'config': {},
            'additional_option': None

        }
        with self.assertRaisesRegex(ValidationError,
                                    r"Additional properties are not allowed \('additional_option' was unexpected\)"):
            validate(v, effects_jsonschema_stub)


class TestPropertyValidatorClass(SixTestCase):
    def test_propertyvalidator_produces_valid_json(self):
        pv = PropertyValidator()
        pv.validate_json()


class TestValidatorEffects(SixTestCase):
    def test_validator_effect_produces_valid_json(self):
        ve = ValidatorEffect()
        ve.validate_json()

    def test_visual_effect_produces_valid_json(self):
        ve = VisualEffect()
        self.assertTrue('applyCss' in ve.as_json().get('config'))
        ve.validate_json()

    def test_valid_visualeffect_produces_valid_json(self):
        ve = ValidVisualEffect()
        self.assertTrue('applyCss' in ve.as_json().get('config'))
        ve.validate_json()

    def test_invalid_visualeffect_productes_valid_json(self):
        ve = InvalidVisualEffect()
        self.assertTrue('applyCss' in ve.as_json().get('config'))
        ve.validate_json()


class TestValidatorParsing(SixTestCase):

    def test_valid_numeric_range_validator_json(self):
        validator_json = dict(
            vtype='numericRangeValidator',
            config=dict(
                minvalue=2,
                maxvalue=10,
                stepsize=2,
                enforce_stepsize=False,
                on_valid=[dict(effect="visualEffect", config=dict(applyCss="valid"))],
                on_invalid=[dict(effect="visualEffect", config=dict(applyCss="invalid")),
                            dict(effect="errorTextEffect",
                                 config=dict(text="Range should be between 2 and 10 with step 2."))]
            )
        )

        validator = PropertyValidator.parse(validator_json)

        self.assertIsInstance(validator, NumericRangeValidator)
        self.assertTrue(validator.validate_json)
        pass


class TestValidatorDumping(SixTestCase):

    def test_valid_numeric_range_validator_dumped(self):
        validator_json = dict(
            vtype='numericRangeValidator',
            config=dict(
                minvalue=2,
                maxvalue=10,
                stepsize=2,
                enforce_stepsize=False,
                on_valid=[dict(effect="visualEffect", config=dict(applyCss="valid"))],
                on_invalid=[dict(effect="visualEffect", config=dict(applyCss="invalid")),
                            dict(effect="errorTextEffect",
                                 config=dict(text="Range should be between 2 and 10 with step 2."))]
            )
        )

        validator = PropertyValidator.parse(validator_json)
        self.assertIsInstance(validator, NumericRangeValidator)

        dumped_json = validator.as_json()
        self.assertIsNone(validate(dumped_json, validator_jsonschema_stub))
        self.assertIsNone(validator.validate_json())


class TestNumericRangeValidator(SixTestCase):
    def test_numeric_range_without_settings_validated_json(self):
        validator = NumericRangeValidator()

        as_json = validator.as_json()
        self.assertIsNone(validator.validate_json())
        self.assertIsInstance(as_json, dict)
        self.assertDictEqual(as_json, {'config': {}, 'vtype': 'numericRangeValidator'})

    def test_numeric_range_validates_with_lower_bound(self):
        validator = NumericRangeValidator(minvalue=0)
        self.assertTrue(validator.is_valid(1))
        validator = NumericRangeValidator(minvalue=5)
        self.assertTrue(validator.is_valid(6))

    def test_numeric_range_validates_with_max_bound(self):
        validator = NumericRangeValidator(maxvalue=0)
        self.assertTrue(validator.is_valid(-3))
        validator = NumericRangeValidator(maxvalue=10)
        self.assertTrue(validator.is_valid(1))

    def test_numeric_range_validates_with_min_max_bound(self):
        validator = NumericRangeValidator(maxvalue=10, minvalue=0)
        self.assertTrue(validator.is_valid(1))

    def test_numeric_range_invalidates_with_lower_bound(self):
        validator = NumericRangeValidator(minvalue=0)
        self.assertFalse(validator.is_valid(-1))
        validator = NumericRangeValidator(minvalue=5)
        self.assertFalse(validator.is_valid(2))

    def test_numeric_range_invalidates_with_max_bound(self):
        validator = NumericRangeValidator(maxvalue=0)
        self.assertFalse(validator.is_valid(1))
        validator = NumericRangeValidator(maxvalue=5)
        self.assertFalse(validator.is_valid(6))

    def test_numeric_range_invalidates_with_min_max_bound(self):
        validator = NumericRangeValidator(maxvalue=10, minvalue=0)
        self.assertFalse(validator.is_valid(-1))

    def test_numeric_range_validates_with_min_max_bound_float(self):
        validator = NumericRangeValidator(maxvalue=10., minvalue=0.)
        self.assertTrue(validator.is_valid(1.5))

    def test_numeric_range_validates_with_stepsize_float(self):
        validator = NumericRangeValidator(stepsize=0.2, enforce_stepsize=True)
        self.assertTrue(validator.is_valid(1.2))
        self.assertTrue(validator.is_invalid(1.3))

    def test_numeric_range_validates_with_stepsize_int_with_minvalue(self):
        validator = NumericRangeValidator(minvalue=11, stepsize=2, enforce_stepsize=True)
        self.assertTrue(validator.is_valid(13))
        self.assertTrue(validator.is_invalid(16))

    def test_numeric_range_raises_exception_when_min_is_greater_than_max(self):
        with self.assertRaisesRegex(Exception, 'should be smaller than the maxvalue'):
            NumericRangeValidator(minvalue=11, maxvalue=-11)

    def test_numeric_range_raises_exception_when_enforce_stepsize_without_stepsize(self):
        with self.assertRaisesRegex(Exception, 'The stepsize should be provided when enforcing stepsize'):
            NumericRangeValidator(stepsize=None, enforce_stepsize=True)

    def test_numeric_range_does_not_respect_max_issue_435(self):
        """With reference to issue #435

        In KE-chain I've added a property with numeric range validators min = unbound and max = 1000.
        A value of 3333 simply returns as valid in pykechain:
        _validation_reasons =>'Value \'3333\' is between -inf and inf'
        """
        validator = NumericRangeValidator(maxvalue=1000)
        self.assertFalse(validator.is_valid(3333))
        self.assertRegex(validator.get_reason(), "Value '3333' should be between -inf and 1000")


class TestBooleanFieldValidator(SixTestCase):
    def test_boolean_validator_without_settings(self):
        validator = BooleanFieldValidator()
        self.assertIsNone(validator.validate_json())
        self.assertIsInstance(validator.as_json(), dict)
        self.assertDictEqual(validator.as_json(), {'config': {}, 'vtype': 'booleanFieldValidator'})


class TestRequiredFieldValidator(SixTestCase):
    def test_requiredfield_validator_without_settings(self):
        validator = RequiredFieldValidator()
        self.assertIsNone(validator.validate_json())
        self.assertIsInstance(validator.as_json(), dict)
        self.assertDictEqual(validator.as_json(), {'config': {}, 'vtype': 'requiredFieldValidator'})

    def test_requiredfield_validator_is_false_on_nonevalue(self):
        validator = RequiredFieldValidator()
        self.assertFalse(validator.is_valid(None))
        self.assertFalse(validator.is_valid(''))
        self.assertFalse(validator.is_valid([]))
        self.assertFalse(validator.is_valid(list()))
        self.assertFalse(validator.is_valid(tuple()))
        self.assertFalse(validator.is_valid(set()))

        # the value 'False' is not a None value so that is True (placed in other test)
        # self.assertFalse(validator.is_valid(False))

    def test_requiredfield_validator_is_true_on_value(self):
        validator = RequiredFieldValidator()
        self.assertTrue(validator.is_valid(1))
        self.assertTrue(validator.is_valid(0))
        self.assertTrue(validator.is_valid(0.0))
        self.assertTrue(validator.is_valid(float('inf')))
        self.assertTrue(validator.is_valid('string'))
        self.assertTrue(validator.is_valid({'key': 'val'}))
        self.assertTrue(validator.is_valid([1, 2, 3]))
        self.assertTrue(validator.is_valid((1,)))
        self.assertTrue(validator.is_valid({1, 2, 3}))
        self.assertTrue(validator.is_valid(False))


class TestRegexValidator(SixTestCase):
    def test_regex_validator_without_settings(self):
        validator = RegexStringValidator()
        self.assertIsNone(validator.validate_json())
        self.assertIsInstance(validator.as_json(), dict)
        self.assertDictEqual(validator.as_json(), {'config': {}, 'vtype': 'regexStringValidator'})

    def test_regex_validator_with_pattern_match(self):
        validator = RegexStringValidator(pattern=r'.*')

        self.assertTrue(validator('mr cactus is tevree'))

    def test_regex_validator_without_pattern_match(self):
        validator = RegexStringValidator()

        # per default the regex string matches everything
        self.assertEqual(validator.pattern, '.+')
        self.assertTrue(validator('mr cactus is tevree'))
        self.assertFalse(validator(''))

    def test_regex_validator_fails_on_none_value(self):
        validator = RegexStringValidator(pattern=r'.*')

        self.assertFalse(validator.is_valid(None))

    def test_regex_validator_complex_email_regex(self):
        validator = EmailValidator()

        self.assertTrue(validator.is_valid('support@ke-works.com'))
        self.assertFalse(validator.is_valid('___'))
        self.assertIsNone(validator.is_valid(None))
        self.assertFalse(validator.is_valid('user@domain'))


class TestOddEvenNumberValidator(SixTestCase):
    def test_even_number_validator_without_settings(self):
        validator = EvenNumberValidator()
        self.assertIsNone(validator.validate_json())
        self.assertIsInstance(validator.as_json(), dict)
        self.assertDictEqual(validator.as_json(), {'config': {}, 'vtype': 'evenNumberValidator'})

    def test_odd_number_validator_without_settings(self):
        validator = OddNumberValidator()
        self.assertIsNone(validator.validate_json())
        self.assertIsInstance(validator.as_json(), dict)
        self.assertDictEqual(validator.as_json(), {'config': {}, 'vtype': 'oddNumberValidator'})

    def test_even_number_validator_is_valid(self):
        validator = EvenNumberValidator()
        self.assertTrue(validator.is_valid(2))

    def test_odd_number_validator_is_valid(self):
        validator = OddNumberValidator()
        self.assertTrue(validator.is_valid(3))

    def test_even_number_validator_is_invalid(self):
        validator = EvenNumberValidator()
        self.assertFalse(validator.is_valid(3))

    def test_odd_number_validator_is_invalid(self):
        validator = OddNumberValidator()
        self.assertFalse(validator.is_valid(4))

    def test_even_number_validator_is_none(self):
        validator = EvenNumberValidator()
        self.assertIsNone(validator.is_valid(None))

    def test_odd_number_validator_is_none(self):
        validator = OddNumberValidator()
        self.assertIsNone(validator.is_valid(None))

    def test_even_number_validator_float_valid(self):
        validator = EvenNumberValidator()
        self.assertTrue(validator.is_valid(4.5))
        self.assertTrue(validator.is_valid(4.0))
        self.assertTrue(validator.is_valid(3.9999999999999999))  # rounding accuracy

    def test_odd_number_validator_float_valid(self):
        validator = OddNumberValidator()
        self.assertTrue(validator.is_valid(3.0))
        self.assertTrue(validator.is_valid(4.9999999999999999))  # rounding accuracy

    def test_even_number_validator_float_invalid(self):
        validator = EvenNumberValidator()
        self.assertFalse(validator.is_valid(3.99))
        self.assertFalse(validator.is_valid(5.5))

    def test_odd_number_validator_float_invalid(self):
        validator = OddNumberValidator()
        self.assertFalse(validator.is_valid(4.99))
        self.assertFalse(validator.is_valid(6.0))


    def test_even_number_validator_invalid_input(self):
        validator = EvenNumberValidator()
        self.assertFalse(validator.is_valid(dict()))
        self.assertFalse(validator.is_valid(list))
        self.assertFalse(validator.is_valid(set()))
        self.assertFalse(validator.is_valid(tuple()))
        self.assertFalse(validator.is_valid("3"))

    def test_odd_number_validator_invalid_input(self):
        validator = OddNumberValidator()
        self.assertFalse(validator.is_valid(dict()))
        self.assertFalse(validator.is_valid(list))
        self.assertFalse(validator.is_valid(set()))
        self.assertFalse(validator.is_valid(tuple()))
        self.assertFalse(validator.is_valid("3"))


class TestSingleReferenceValidator(SixTestCase):
    def test_singlereference_validator_without_settings(self):
        validator = SingleReferenceValidator()
        self.assertIsNone(validator.validate_json())
        self.assertIsInstance(validator.as_json(), dict)
        self.assertDictEqual(validator.as_json(), {'config': {}, 'vtype': 'singleReferenceValidator'})

    def test_singlereference_validator_is_valid(self):
        validator = SingleReferenceValidator()
        self.assertIsNone(validator.is_valid(None))
        self.assertTrue(validator.is_valid(list()))
        self.assertTrue(validator.is_valid(set()))
        self.assertTrue(validator.is_valid(tuple()))
        self.assertTrue(validator.is_valid(('first selection',)))

    def test_singlereference_validator_is_invalid(self):
        validator = SingleReferenceValidator()
        self.assertFalse(validator.is_valid(["first", "second"]))
        self.assertFalse(validator.is_valid((1, 2)))

    def test_singlerefence_validator_is_invalid_with_invalid_values(self):
        validator = SingleReferenceValidator()
        self.assertFalse(validator.is_valid("a string"))
        self.assertFalse(validator.is_valid(1.0))
        self.assertFalse(validator.is_valid(dict()))


class TestAlwaysAllowValidator(SixTestCase):
    def test_always_allow_validator_without_settings(self):
        validator = AlwaysAllowValidator()
        self.assertTrue(validator.is_valid('some text'))
        self.assertTrue(validator.is_valid(42))
        self.assertTrue(validator.is_valid(42.0))
        self.assertTrue(validator.is_valid(['a', 'b']))
        self.assertTrue(validator.is_valid(True))
        self.assertTrue(validator.is_valid(False))
        self.assertTrue(validator.is_valid(None))
        self.assertTrue(validator.is_valid({"a":"b"}))
        self.assertFalse(validator.is_invalid(None))
        self.assertTrue(validator.is_valid(list()))
        self.assertTrue(validator.is_valid(set()))
        self.assertTrue(validator.is_valid(tuple()))
        self.assertIsNone(validator.validate_json())


class TestFileSizeValidator(SixTestCase):
    def test_validator_valid_json_with_settings(self):
        validator = FileSizeValidator(max_size=100)
        self.assertIsNone(validator.validate_json())

    def test_validator_valid_json_without_settings(self):
        validator = FileSizeValidator()
        self.assertIsNone(validator.validate_json())

    def test_validator_valid_json_with_additional_arguments(self):
        validator = FileSizeValidator(max_size=100, another_argument=False)
        # another_argument is not found in the json
        validator.validate_json()
        validator.as_json()

    def test_validator_invalid_arguments(self):
        with self.assertRaisesRegex(ValueError, 'should be a number'):
            FileSizeValidator(max_size=set())
        with self.assertRaisesRegex(ValueError, 'should be a number'):
            FileSizeValidator(max_size=list())
        with self.assertRaisesRegex(ValueError, 'should be a number'):
            FileSizeValidator(max_size=dict())
        with self.assertRaisesRegex(ValueError, 'should be a number'):
            FileSizeValidator(max_size=tuple())
        with self.assertRaisesRegex(ValueError, 'should be a number'):
            FileSizeValidator(max_size=tuple())
        with self.assertRaisesRegex(ValueError, 'should be a number'):
            FileSizeValidator(max_size='100')

    def test_filesizevalidator_being_valid(self):
        validator = FileSizeValidator(max_size=100)

        self.assertTrue(validator.is_valid(100))
        self.assertTrue(validator.is_valid(100.))
        self.assertTrue(validator.is_valid(99))
        self.assertTrue(validator.is_valid(0))

        self.assertTrue(validator.is_valid('text'))
        self.assertTrue(validator.is_valid(list()))
        self.assertTrue(validator.is_valid(tuple()))
        self.assertTrue(validator.is_valid(dict()))
        self.assertTrue(validator.is_valid(set()))

    def test_filesizevalidator_being_invalid(self):
        validator = FileSizeValidator(max_size=100)

        self.assertFalse(validator.is_valid(101))
        self.assertFalse(validator.is_valid(101.))
        self.assertFalse(validator.is_valid(-1))

    def test_filesizevalidator_being_none(self):
        validator = FileSizeValidator(max_size=100)

        self.assertIsNone(validator.is_valid(None))


class TestFileExtensionValidator(SixTestCase):
    def test_validator_valid_json_with_settings(self):
        validator = FileExtensionValidator(accept=[".csv"])
        self.assertIsNone(validator.validate_json())

    def test_validator_valid_json_without_settings(self):
        validator = FileExtensionValidator()
        self.assertIsNone(validator.validate_json())

    def test_fileextensionvalidator_on_extension(self):
        validator = FileExtensionValidator(accept=[".csv"])
        self.assertTrue(validator.is_valid('file.csv'))
        self.assertFalse(validator.is_valid('file.tsv'))
        self.assertFalse(validator.is_valid('file.txt'))

    def test_fileextensionvalidator_on_extensions(self):
        validator = FileExtensionValidator(accept=[".csv", ".pdf"])
        self.assertTrue(validator.is_valid('file.csv'))
        self.assertTrue(validator.is_valid('file.pdf'))
        self.assertFalse(validator.is_valid('file.docx'))

    def test_fileextensionvalidator_on_mimetype(self):
        validator = FileExtensionValidator(accept=["application/pdf"])
        self.assertTrue(validator.is_valid('file.pdf'))
        self.assertFalse(validator.is_valid('file.docx'))

    def test_fileextensionvalidator_on_expanded_mimetypes(self):
        validator = FileExtensionValidator(accept=["video/*"])
        self.assertTrue(validator.is_valid('file.mp4'))
        self.assertTrue(validator.is_valid('file.mov'))
        self.assertTrue(validator.is_valid('file.mpeg'))
        self.assertTrue(validator.is_valid('file.flv'))
        self.assertTrue(validator.is_valid('file.wmv'))
        self.assertTrue(validator.is_valid('file.webm'))

    def test_fileextensionvalidator_on_mixed_mode(self):
        validator = FileExtensionValidator(accept=[".csv", "application/pdf"])
        self.assertTrue(validator.is_valid('file.csv'))
        self.assertTrue(validator.is_valid('file.pdf'))
        self.assertFalse(validator.is_valid('file.docx'))

    def test_fileextensionvalidator_on_excel(self):
        validator = FileExtensionValidator(accept=[".xls", ".xlsx"])
        self.assertTrue(validator.is_valid('file.xls'))
        self.assertTrue(validator.is_valid('file.xlsx'))
        self.assertFalse(validator.is_valid('file.csv'))

    def test_fileextensionvalidator_being_none(self):
        validator = FileExtensionValidator(accept=["video/*"])

        self.assertIsNone(validator.is_valid(None))

class TestPropertyWithValidator(SixTestCase):

    def test_property_without_validator(self):
        prop = Property2(json={}, client=None)
        self.assertIsNone(prop.is_valid)
        self.assertIsNone(prop.is_invalid)
        self.assertEqual(prop._validators, list())
        self.assertEqual(prop.validate(), list())

    def test_property_with_numeric_range_validator(self):
        prop_json = dict(
            value=1,
            value_options=dict(
                validators=[NumericRangeValidator(minvalue=0, maxvalue=10).as_json()]
            ))
        prop = Property2(json=prop_json, client=None)
        self.assertTrue(prop.is_valid)
        self.assertFalse(prop.is_invalid)
        self.assertTrue(prop.validate())

    def test_property_with_numeric_range_validator_value_is_none(self):
        prop_json = dict(
            value=None,
            value_options=dict(
                validators=[NumericRangeValidator(minvalue=0, maxvalue=10).as_json()]
            ))
        prop = Property2(json=prop_json, client=None)
        self.assertIsNone(prop.is_valid)
        self.assertIsNone(prop.is_invalid)
        self.assertListEqual([(None, "No reason")], prop.validate())


    def test_property_with_filesize_validator(self):
        prop_json = dict(
            value="attachments/12345678-1234-5678-1234-567812345678/some_file.txt",
            value_options=dict(
                validators=[FileSizeValidator(max_size=100).as_json()]
            ))
        prop = AttachmentProperty2(json=prop_json, client=None)
        self.assertTrue(prop.is_valid)
        self.assertFalse(prop.is_invalid)
        self.assertTrue(prop.validate())

    def test_property_without_value_with_filesize_validator(self):
        prop_json = dict(
            value=None,
            value_options=dict(
                validators=[FileSizeValidator(max_size=100).as_json()]
            ))
        prop = AttachmentProperty2(json=prop_json, client=None)
        self.assertIsNone(prop.is_valid)
        self.assertIsNone(prop.is_invalid)
        self.assertListEqual([(None, "No reason")], prop.validate())

    def test_property_with_fileextension_validator(self):
        prop_json = dict(
            value="attachments/12345678-1234-5678-1234-567812345678/some_file.txt",
            value_options=dict(
                validators=[FileExtensionValidator(accept=['.txt', 'application/pdf']).as_json()]
            ))
        prop = AttachmentProperty2(json=prop_json, client=None)
        self.assertTrue(prop.is_valid)
        self.assertFalse(prop.is_invalid)
        self.assertTrue(prop.validate())

    def test_property_without_value_with_fileextension_validator(self):
        prop_json = dict(
            value=None,
            value_options=dict(
                validators=[FileExtensionValidator(accept=['.txt', 'application/pdf']).as_json()]
            ))
        prop = AttachmentProperty2(json=prop_json, client=None)
        self.assertIsNone(prop.is_valid)
        self.assertIsNone(prop.is_invalid)
        self.assertListEqual([(None, "No reason")], prop.validate())

class TestPropertyWithValidatorFromLiveServer(TestBetamax):

    def setUp(self):
        super(TestPropertyWithValidatorFromLiveServer, self).setUp()
        self.part_model = self.project.model(name='Model')
        self.numeric_range_prop_model = self.part_model.add_property(name='numericrange_validatortest',
                                                                     property_type=PropertyType.FLOAT_VALUE)
        self.numeric_range_prop_model.validators = [NumericRangeValidator(min=0., max=42.)]
        self.part = self.project.part(name__startswith='Catalog').add(self.part_model, name='Model Instance for tests')
        self.numeric_range_prop_instance = self.part.property(name='numericrange_validatortest')

    def tearDown(self):
        self.numeric_range_prop_model.delete()
        self.part.delete()
        super(TestPropertyWithValidatorFromLiveServer, self).tearDown()

    def test_numeric_property_with_validator_parses(self):
        self.assertIsInstance(self.numeric_range_prop_instance._validators, list)
        self.assertIsInstance(self.numeric_range_prop_instance._validators[0], PropertyValidator)

    def test_numeric_property_add_requiredvalidator_on_model(self):
        # test
        validators = self.numeric_range_prop_model.validators
        validators.append(RequiredFieldValidator())
        self.numeric_range_prop_model.validators = validators

        for validator in self.numeric_range_prop_model.validators:
            self.assertIsInstance(validator, (NumericRangeValidator, RequiredFieldValidator))

    def test_numeric_property_add_requiredvalidator_on_instance(self):
        # test
        validators = self.numeric_range_prop_instance.validators
        validators.append(RequiredFieldValidator())

        with self.assertRaisesRegex(IllegalArgumentError, "To update the list of validators, it can only work "
                                                          "on `Property` of category 'MODEL'"):
            self.numeric_range_prop_instance.validators = validators



from datetime import datetime

from pykechain.enums import PropertyType, Category
from pykechain.exceptions import NotFoundError, APIError, IllegalArgumentError
from pykechain.models import Property
from pykechain.models.validators import SingleReferenceValidator
from tests.classes import TestBetamax


class TestProperties(TestBetamax):
    def setUp(self):
        super(TestProperties, self).setUp()

        self.wheel_model = self.project.model('Wheel')
        self.bike = self.project.model('Bike')
        self.test_property_model = self.bike.add_property(name="__Test property @ {}".format(datetime.now()),
                                                          property_type=PropertyType.INT_VALUE,
                                                          description='Description of test property',
                                                          unit='units',
                                                          default_value=42)
        self.front_wheel = self.project.part(name='Front Wheel')
        self.test_property_instance = self.bike.instance().property(name=self.test_property_model.name)

        # single_reference_validator = SingleReferenceValidator()
        # self.test_ref_property_model = self.bike.add_property(name="__Test ref property @ {}".format(datetime.now()),
        #                                                       property_type=PropertyType.REFERENCES_VALUE,
        #                                                       description='Description of test ref property',
        #                                                       unit='no unit',
        #                                                       default_value=[self.wheel_model])
        # self.test_ref_property_model.validators = [single_reference_validator]

    def tearDown(self):
        self.test_property_model.delete()
        super(TestProperties, self).tearDown()

    def test_retrieve_properties(self):
        properties = self.project.properties('Diameter')

        self.assertTrue(len(properties) > 0)

    def test_retrieve_properties_with_kwargs(self):
        # setUp
        bike = self.project.part(name='Bike')
        properties_with_kwargs = self.client.properties(part_id=bike.id)

        self.assertTrue(properties_with_kwargs)

        # testing
        for prop in properties_with_kwargs:
            self.assertEqual(prop.part.id, bike.id)

    def test_retrieve_property(self):
        prop = self.project.property(name='Cheap?', category=Category.MODEL)

        self.assertTrue(prop)

    def test_property_attributes(self):
        attributes = ['_client', '_json_data', 'id', 'name', 'created_at', 'updated_at', 'ref',
                      '_output', '_value', '_options', 'type', 'category',
                      '_validators']

        obj = self.project.property(name='Cheap?', category=Category.MODEL)
        for attribute in attributes:
            self.assertTrue(hasattr(obj, attribute),
                            "Could not find '{}' in the object: '{}'".format(attribute, obj.__dict__.keys()))

    def test_get_property_by_name(self):
        bike = self.project.part('Bike')
        # retrieve the property Gears directly via an API call
        gears_property = self.project.properties(name='Gears', category=Category.INSTANCE)[0]

        self.assertEqual(bike.property('Gears'), gears_property)

    def test_get_property_by_uuid(self):
        bike = self.project.part('Bike')
        gears_id = bike.property('Gears').id
        # retrieve the property Gears directly via an API call
        gears_property = self.project.properties(pk=gears_id, category=Category.INSTANCE)[0]

        self.assertEqual(bike.property(gears_id), gears_property)

    def test_get_invalid_property(self):
        bike = self.project.part('Bike')

        with self.assertRaises(NotFoundError):
            bike.property('Price')

    def test_set_property(self):
        gears = self.project.part('Bike').property('Gears')

        gears.value = 5

        self.assertEqual(gears.value, 5)

        gears.value = 2

        self.assertEqual(self.project.part('Bike').property('Gears').value, 2)

        gears.value = 10

    def test_property_to_part(self):
        bike = self.project.part('Bike')

        bike2 = bike.property('Gears').part

        self.assertEqual(bike.id, bike2.id)

    def test_create_and_delete_property_model(self):
        # Retrieve the bike model
        bike = self.project.model('Bike')

        # test creation of new property model of bike
        new_property = self.client.create_property(model=bike, name='New property', description='Nice prop',
                                                   property_type=PropertyType.CHAR_VALUE, default_value='EUREKA!')

        # check whether the property has been created and whether it's name and type are correct
        self.assertIsInstance(bike.property('New property'), Property)
        self.assertEqual(bike.property('New property').value, 'EUREKA!')
        self.assertEqual(bike.property('New property')._json_data['property_type'], 'CHAR_VALUE')

        # Now delete the property model
        new_property.delete()

        # Retrieve the bike model again
        updated_bike = self.project.model('Bike')

        # Check whether it still has the property model that has just been deleted
        with self.assertRaises(NotFoundError):
            updated_bike.property('New property')

    def test_create_property_where_model_is_instance(self):
        # setUp
        bike_instance = self.project.part(name='Bike')

        # testing
        with self.assertRaises(IllegalArgumentError):
            self.client.create_property(name='Properties are created on models only', model=bike_instance)

    def test_wrongly_creation_of_property(self):
        # These actions should not be possible. This test is of course, expecting APIErrors to be raised
        bike_model = self.project.model('Bike')

        # Test whether an integer property model can be created with an incorrect default value
        with self.assertRaises(APIError):
            bike_model.add_property(name='Integer',
                                    property_type=PropertyType.INT_VALUE,
                                    default_value='Why is there a string here?')

    # 1.7.2
    def test_get_partmodel_of_propertymodel(self):
        """As found by @joost.schut, see #119 - thanks!"""

        wheel_model = self.project.model('Wheel')
        spokes_model = wheel_model.property('Spokes')
        part_of_spokes_model = spokes_model.part

        self.assertTrue(wheel_model.id == part_of_spokes_model.id)

    # 1.11
    def test_edit_property_model_name(self):
        bike_model = self.project.model('Bike')
        gears_property = bike_model.property(name='Gears')
        gears_old_name = gears_property.name
        gears_property.edit(name='Cogs')
        gears_property_u = bike_model.property(name='Cogs')

        self.assertEqual(gears_property.id, gears_property_u.id)
        self.assertEqual(gears_property.name, gears_property_u.name)
        self.assertEqual(gears_property.name, 'Cogs')

        with self.assertRaises(IllegalArgumentError):
            gears_property.edit(name=True)
        # teardown
        gears_property.edit(name=gears_old_name)

    def test_edit_property_model_description(self):
        bike_model = self.project.model('Bike')
        gears_property = bike_model.property(name='Gears')
        gears_old_description = str(gears_property._json_data.get('description'))

        new_description = 'Cogs, definitely cogs.'
        gears_property.edit(description=new_description)
        gears_property_u = bike_model.property(name='Gears')

        self.assertEqual(gears_property.id, gears_property_u.id)
        with self.assertRaises(IllegalArgumentError):
            gears_property.edit(description=True)

        # teardown
        gears_property.edit(description=gears_old_description)

    def test_edit_property_model_unit(self):
        front_fork_model = self.project.model('Front Fork')
        height_property = front_fork_model.property(name='Height')
        height_old_unit = str(height_property._json_data.get('unit'))
        new_unit = 'm'

        height_property.edit(unit=new_unit)

        height_property_u = front_fork_model.property(name='Height')
        self.assertEqual(height_property.id, height_property_u.id)

        with self.assertRaises(IllegalArgumentError):
            height_property.edit(unit=4)

        # teardown
        height_property.edit(unit=height_old_unit)

    # 1.11
    def test_set_property_for_real(self):
        gears = self.project.part('Bike').property('Gears')
        old_value = int(gears.value)

        # set the value
        new_value = old_value * 20 + 1
        self.assertNotEqual(new_value, old_value)

        gears.value = new_value
        self.assertEqual(gears.value, new_value)
        self.assertEqual(gears._value, new_value)
        self.assertEqual(gears._json_data.get('value'), new_value)

        # teardown
        gears.value = old_value

    # 1.16
    def test_creation_of_all_property_model_types(self):
        # set up
        bike_model = self.project.model(name='Bike')

        # create a property model for each property type
        no_value_property_type = bike_model.add_property(name='Not correct property type',
                                                         property_type=PropertyType.CHAR_VALUE)
        no_property_type_specified = bike_model.add_property(name='No property type specified')
        single_line_text = bike_model.add_property(name='Single line text', property_type=PropertyType.CHAR_VALUE)
        multi_line_text = bike_model.add_property(name='Multi line text', property_type=PropertyType.TEXT_VALUE)
        integer = bike_model.add_property(name='Integer', property_type=PropertyType.INT_VALUE)
        decimal = bike_model.add_property(name='Float', property_type=PropertyType.FLOAT_VALUE)
        boolean = bike_model.add_property(name='Boolean', property_type=PropertyType.BOOLEAN_VALUE)
        datetime = bike_model.add_property(name='Datetime', property_type=PropertyType.DATETIME_VALUE)
        attachment = bike_model.add_property(name='Attachment', property_type=PropertyType.ATTACHMENT_VALUE)
        link = bike_model.add_property(name='Link', property_type=PropertyType.LINK_VALUE)
        single_select = bike_model.add_property(name='Single select', property_type=PropertyType.SINGLE_SELECT_VALUE)
        reference = bike_model.add_property(name='Reference', property_type=PropertyType.REFERENCES_VALUE)

        # teardown
        no_value_property_type.delete()
        no_property_type_specified.delete()
        single_line_text.delete()
        multi_line_text.delete()
        integer.delete()
        decimal.delete()
        boolean.delete()
        datetime.delete()
        attachment.delete()
        link.delete()
        single_select.delete()
        reference.delete()

    # 2.5.4
    def test_property_type_property(self):
        # set up
        front_fork_model = self.project.model('Front Fork')
        height_property = front_fork_model.property(name='Height')

        self.assertEqual(str(height_property.type), PropertyType.FLOAT_VALUE)

    def test_property_unit_property(self):
        # set up
        front_fork_model = self.project.model('Front Fork')
        height_property = front_fork_model.property(name='Height')

        self.assertEqual(str(height_property.unit), str(height_property._json_data.get('unit')))

    def test_property_description_property(self):
        # set up
        front_fork_model = self.project.model('Front Fork')
        height_property = front_fork_model.property(name='Height')

        self.assertEqual(str(height_property.description), str(height_property._json_data.get('description')))

    # 3.0
    def test_copy_property_model(self):
        # setUp
        copied_property = self.test_property_model.copy(target_part=self.wheel_model, name='Copied property')

        # testing
        self.assertEqual(copied_property.name, 'Copied property')
        self.assertEqual(copied_property.description, self.test_property_model.description)
        self.assertEqual(copied_property.unit, self.test_property_model.unit)
        self.assertEqual(copied_property.value, self.test_property_model.value)

        # tearDown
        copied_property.delete()

    def test_move_property_model(self):
        # setUp
        moved_property = self.test_property_model.move(target_part=self.wheel_model, name='Copied property')

        # testing
        self.assertEqual(moved_property.name, 'Copied property')
        self.assertEqual(moved_property.description, self.test_property_model.description)
        self.assertEqual(moved_property.unit, self.test_property_model.unit)
        self.assertEqual(moved_property.value, self.test_property_model.value)
        with self.assertRaises(NotFoundError):
            self.project.property(id=self.test_property_model.id)

        # tearDown
        # for a correct teardown in the unittest we need to -reassign- the moved one to the self.test_property_model
        self.test_property_model = moved_property

    def test_copy_property_instance(self):
        # setUp
        self.test_property_instance.value = 200
        copied_property = self.test_property_instance.copy(target_part=self.front_wheel,
                                                           name='Copied property instance')

        # testing
        self.assertEqual(copied_property.name, 'Copied property instance')
        self.assertEqual(copied_property.description, self.test_property_instance.description)
        self.assertEqual(copied_property.unit, self.test_property_instance.unit)
        self.assertEqual(copied_property.value, self.test_property_instance.value)

        # tearDown
        copied_property.model().delete()




class TestPropertiesWithReferenceProperty(TestBetamax):
    def setUp(self):
        super(TestPropertiesWithReferenceProperty, self).setUp()

        self.wheel_model = self.project.model('Wheel')
        self.bike = self.project.model('Bike')


        self.test_ref_property_model = self.bike.add_property(name="__Test ref property @ {}".format(datetime.now()),
                                                              property_type=PropertyType.REFERENCES_VALUE,
                                                              description='Description of test ref property',
                                                              unit='no unit',
                                                              default_value=[self.wheel_model])
        self.test_ref_property_model.validators = [SingleReferenceValidator()]

    def tearDown(self):
        self.test_ref_property_model.delete()
        super(TestPropertiesWithReferenceProperty, self).tearDown()

    def test_copy_reference_property_with_options(self):
        # setUp
        copied_ref_property = self.test_ref_property_model.copy(target_part=self.wheel_model,
                                                                name='__Copied ref property')

        # testing
        self.assertEqual(copied_ref_property.name, '__Copied ref property')
        self.assertEqual(copied_ref_property.description, self.test_ref_property_model.description)
        self.assertEqual(copied_ref_property.unit, self.test_ref_property_model.unit)
        self.assertEqual(copied_ref_property.value, self.test_ref_property_model.value)
        self.assertEqual(copied_ref_property._options, self.test_ref_property_model._options)

        # tearDown
        copied_ref_property.delete()



from pykechain.enums import WidgetTypes, ShowColumnTypes, NavigationBarAlignment, FilterType, ProgressBarColors, \
    Category, LinkTargets
from pykechain.exceptions import IllegalArgumentError
from pykechain.models import Activity, Activity2
from pykechain.models.widgets import UndefinedWidget, HtmlWidget, PropertygridWidget, AttachmentviewerWidget, \
    SupergridWidget, FilteredgridWidget, TasknavigationbarWidget, SignatureWidget, ServiceWidget, NotebookWidget, \
    MulticolumnWidget, CardWidget, MetapanelWidget
from pykechain.models.widgets.widget import Widget
from pykechain.models.widgets.widgets_manager import WidgetsManager
from pykechain.utils import is_uuid
from tests.classes import TestBetamax


class TestWidgets(TestBetamax):

    def setUp(self):
        super(TestWidgets, self).setUp()

    def test_retrieve_widgets_in_activity(self):
        activity = self.project.activity('Task - Form + Tables + Service')
        widget_set = activity.widgets()
        self.assertIsInstance(widget_set, WidgetsManager)
        for w in widget_set:
            self.assertIsInstance(w, Widget)

    def test_create_widget_in_activity(self):
        activity = self.project.activity('test task')  # type: Activity
        created_widget = self.client.create_widget(
            title="Test_widget",
            activity=activity,
            widget_type=WidgetTypes.UNDEFINED,
            meta={},
            order=100
        )
        self.assertIsInstance(created_widget, UndefinedWidget)
        created_widget.delete()

    def test_widget_attributes(self):
        attributes = ['_client', '_json_data', 'id', 'created_at', 'updated_at', 'ref',
                      'widget_type', 'title', 'meta', 'order', '_activity_id', '_parent_id',
                      'has_subwidgets', '_scope_id', 'progress']

        obj = self.project.activity('Specify wheel diameter').widgets()[0]
        self.assertIsInstance(obj, Widget)
        for attribute in attributes:
            self.assertTrue(hasattr(obj, attribute),
                            "Could not find '{}' in the object: '{}'".format(attribute, obj.__dict__.keys()))

    def test_widget_meta_attribute_is_not_None(self):
        obj = self.project.activity('Specify wheel diameter').widgets()[0]
        self.assertIsInstance(obj, Widget)
        self.assertIsNotNone(obj.meta)

class TestWidgetManager(TestBetamax):
    def test_activity_has_metapanel_in_widget_manager(self):
        activity = self.project.activity('Task - Form + Tables + Service')
        widgets = activity.widgets()
        self.assertIsInstance(widgets, WidgetsManager)
        self.assertTrue(len(widgets) >= 5)

        self.assertTrue(widgets[0].widget_type, WidgetTypes.METAPANEL)

    def test_widget_can_be_found_with_uuid_in_widget_manager(self):
        activity = self.project.activity('Task - Form + Tables + Service')
        widgets = activity.widgets()
        metapanel = widgets[0]
        self.assertTrue(metapanel, WidgetTypes.METAPANEL)

        self.assertEqual(widgets[metapanel.id], metapanel)

    def test_widgetmanager_has_activity_and_client(self):
        activity = self.project.activity('Task - Form + Tables + Service')
        widgets = activity.widgets()
        self.assertIsNotNone(widgets._client)
        self.assertIsNotNone(widgets._activity_id)
        self.assertIsInstance(widgets._client, self.client.__class__)
        self.assertTrue(is_uuid(widgets._activity_id))


class TestWidgetManagerInActivity(TestBetamax):
    def setUp(self):
        super(TestWidgetManagerInActivity, self).setUp()
        self.task = self.project.create_activity(name="widget_test_task")  # type: Activity2

        self.frame = self.project.part(name='Frame')
        self.frame_model = self.project.model(name='Frame')

    def tearDown(self):
        self.task.delete()
        super(TestWidgetManagerInActivity, self).tearDown()

    def test_new_widget_using_widget_manager(self):
        widget_manager = self.task.widgets()  # type: WidgetsManager

        self.assertEqual(len(widget_manager), 1)
        metapanel = widget_manager[0]

        htmlwidget = widget_manager.create_widget(
            widget_type=WidgetTypes.HTML,
            title="Test HTML widget",
            meta=dict(htmlContent="Hello")
        )

        self.assertIsInstance(metapanel, MetapanelWidget)
        self.assertIsInstance(htmlwidget, HtmlWidget)
        self.assertEqual(len(widget_manager), 1 + 1)

    def test_edit_widget(self):
        widget_manager = self.task.widgets()  # type: WidgetsManager
        bike_part = self.project.part('Bike')
        widget = widget_manager.add_propertygrid_widget(part_instance=bike_part,
                                                        writable_models=bike_part.model().properties)
        import copy
        updated_meta = copy.deepcopy(widget.meta)
        new_title = "My customly edited title"
        updated_meta.update({
            "showTitleValue": "Custom title",
            "customTitle": new_title
        })
        widget_manager[widget.id].edit(meta=updated_meta)

        self.assertEqual(widget_manager[widget.id].meta.get('customTitle'), new_title)

    def test_property_grid_with_associations_using_widget_manager(self):
        widget_manager = self.task.widgets()  # type: WidgetsManager

        widget = widget_manager.create_widget(
            widget_type=WidgetTypes.PROPERTYGRID,
            title="Frame Property Grid",
            meta=dict(
                activityId=str(self.task.id),
                partInstanceId=str(self.frame.id)
            ),
            writable_models=self.frame_model.properties,
            readable_models=[]
        )

        self.assertIsInstance(widget, PropertygridWidget)
        self.assertEqual(len(widget_manager), 1 + 1)

    def test_attachment_widget_with_associations_using_widget_manager(self):
        widget_manager = self.task.widgets()  # type: WidgetsManager
        foto_property = self.project.property("Picture")

        widget = widget_manager.add_attachmentviewer_widget(
            widget_type=WidgetTypes.ATTACHMENTVIEWER,
            title="Attachment Viewer",
            attachment_property=foto_property,
            readable_models=[foto_property.model_id]
        )

        self.assertIsInstance(widget, AttachmentviewerWidget)
        self.assertEqual(len(widget_manager), 1 + 1)

    def test_add_super_grid_widget(self):
        widget_manager = self.task.widgets()  # type: WidgetsManager
        part_model = self.project.model(name='Wheel')
        parent_instance = self.project.part(name='Bike')

        widget = widget_manager.add_supergrid_widget(
            part_model=part_model,
            parent_instance=parent_instance,
            edit=False,
            emphasize_edit=True,
            all_readable=True,
            incomplete_rows=True
        )

        self.assertIsInstance(widget, SupergridWidget)
        self.assertEqual(len(widget_manager), 1 + 1)

    def test_add_filtered_grid_widget(self):
        widget_manager = self.task.widgets()  # type: WidgetsManager
        part_model = self.project.model(name='Wheel')
        parent_instance = self.project.part(name='Bike')
        widget = widget_manager.add_filteredgrid_widget(
            part_model=part_model,
            parent_instance=parent_instance,
            edit=True,
            # sort_property=part_model.property(name='Diameter'),
            emphasize_edit=True,
            all_writable=True,
            collapse_filters=False,
        )

        self.assertIsInstance(widget, FilteredgridWidget)
        self.assertEqual(len(widget_manager), 1 + 1)

    def test_add_filtered_grid_widget_with_prefilters_and_excluded_propmodels(self):
        widget_manager = self.task.widgets()  # type: WidgetsManager
        part_model = self.project.model(name='Wheel')
        parent_instance = self.project.part(name='Bike')
        excluded_propmodels = [part_model.property(name='Spokes')]
        prefilters = dict(
            property_models=[part_model.property(name='Diameter')],
            values=[66],
            filters_type=[FilterType.LOWER_THAN_EQUAL]
        )
        widget = widget_manager.add_filteredgrid_widget(
            part_model=part_model,
            parent_instance=parent_instance,
            edit=True,
            sort_name=True,
            emphasize_edit=True,
            all_writable=True,
            collapse_filters=False,
            prefilters=prefilters,
            excluded_propmodels=excluded_propmodels
        )

        self.assertIsInstance(widget, FilteredgridWidget)
        self.assertEqual(len(widget_manager), 1 + 1)

    def test_add_attachment_widget(self):
        widget_manager = self.task.widgets()
        picture_instance = self.project.part('Bike').property('Picture')
        widget = widget_manager.add_attachmentviewer_widget(
            attachment_property=picture_instance
        )

        self.assertIsInstance(widget, AttachmentviewerWidget)
        self.assertEqual(len(widget_manager), 1 + 1)

    def test_add_propertygrid_widget(self):
        widget_manager = self.task.widgets()  # type: WidgetsManager
        bike_part = self.project.part(name='Bike')
        widget = widget_manager.add_propertygrid_widget(part_instance=bike_part,
                                                        title="Testing the custom title of a property grid widget",
                                                        show_headers=False, show_columns=[ShowColumnTypes.UNIT],
                                                        readable_models=bike_part.model().properties[:2],
                                                        writable_models=bike_part.model().properties[3:])

        self.assertIsInstance(widget, PropertygridWidget)
        self.assertEqual(len(widget_manager), 1 + 1)

    def test_add_signature_widget(self):
        widget_manager = self.task.widgets()  # type: WidgetsManager
        bike_part = self.project.part(name='Bike')
        picture = bike_part.property(name='Picture')

        widget1 = widget_manager.add_signature_widget(attachment_property=picture, title="Yes, my precious",
                                                      custom_undo_button_text="Remove za widget",
                                                      custom_button_text="Sign za widget")
        widget2 = widget_manager.add_signature_widget(attachment_property=picture, title=False,
                                                      custom_undo_button_text=False,
                                                      custom_button_text=False)
        widget3 = widget_manager.add_signature_widget(attachment_property=picture)
        self.assertIsInstance(widget1, SignatureWidget)
        self.assertIsInstance(widget2, SignatureWidget)
        self.assertIsInstance(widget3, SignatureWidget)
        self.assertEqual(len(widget_manager), 1 + 3)

        with self.assertRaises(IllegalArgumentError):
            widget_manager.add_signature_widget(attachment_property='Failed script')

    def test_add_card_widget(self):
        # setUp
        widget_manager = self.task.widgets()  # type: WidgetsManager
        bike_part = self.project.part(name='Bike')
        picture = bike_part.property(name='Picture')
        activity = self.project.activity(name='test task')

        widget1 = widget_manager.add_card_widget(title="Some title", description='Some description',
                                                 link='www.ke-chain.com', link_target=LinkTargets.NEW_TAB)
        widget2 = widget_manager.add_card_widget(image=picture, title=False,
                                                 link=activity.id, link_target=LinkTargets.SAME_TAB)

        # testing
        self.assertEqual(len(widget_manager), 1 + 2)

        self.assertIsInstance(widget1, CardWidget)
        self.assertIsInstance(widget2, CardWidget)

        with self.assertRaises(IllegalArgumentError):
            widget_manager.add_card_widget(title=12)

        with self.assertRaises(IllegalArgumentError):
            widget_manager.add_card_widget(description=bike_part)

        with self.assertRaises(IllegalArgumentError):
            widget_manager.add_card_widget(image='this should not work')

        with self.assertRaises(IllegalArgumentError):
            widget_manager.add_card_widget(link=activity, link_target='_somewhere')

        with self.assertRaises(IllegalArgumentError):
            # noinspection PyTypeChecker
            widget_manager.add_card_widget(image_fit=3)

    def test_service_widget(self):
        widget_manager = self.task.widgets()  # type: WidgetsManager
        service_gears_successful = self.project.service("Service Gears - Successful")

        widget1 = widget_manager.add_service_widget(service=service_gears_successful)
        widget2 = widget_manager.add_service_widget(service=service_gears_successful,
                                                    title=None,
                                                    custom_button_text="Run this script (Custom!) and no title",
                                                    emphasize_run=True,
                                                    download_log=True)
        widget3 = widget_manager.add_service_widget(service=service_gears_successful,
                                                    title="Also a custom title, but no log",
                                                    custom_button_text="Run this script (Custom!)",
                                                    emphasize_run=False,
                                                    show_log=False)

        self.assertIsInstance(widget1, ServiceWidget)
        self.assertIsInstance(widget2, ServiceWidget)
        self.assertIsInstance(widget3, ServiceWidget)
        self.assertEqual(len(widget_manager), 1 + 3)

    def test_add_html_widget(self):
        widget_manager = self.task.widgets()  # type: WidgetsManager
        widget = widget_manager.add_html_widget(html='Or is this just fantasy?',
                                                title='Is this the real life?')

        self.assertIsInstance(widget, HtmlWidget)
        self.assertEqual(len(widget_manager), 1 + 1)

    def test_metapanel_widget(self):
        widget_manager = self.task.widgets()  # type: WidgetsManager
        widget_manager.add_metapanel_widget(
            show_all=True,
            show_progress=True,
            show_progressbar=True
        )
        self.assertEqual(len(widget_manager), 1 + 1)

    def test_add_metapanel_with_progress_settings(self):
        widget_manager = self.task.widgets()  # type: WidgetsManager
        progress_bar = dict(
            height=15,
            showProgressText=False,
            colorCompleted=ProgressBarColors.RED
        )
        widget_manager.add_metapanel_widget(
            show_all=False,
            show_progress=False,
            show_progressbar=True,
            progress_bar=progress_bar
        )
        self.assertEqual(len(widget_manager), 1 + 1)

    def test_progress_widget(self):
        widget_manager = self.task.widgets()
        widget_manager.add_progress_widget(custom_height=35,
                                           show_progress_text=False)
        self.assertEqual(len(widget_manager), 1 + 1)

    def test_delete_all_widgets(self):
        """Delete all widgets from an activity"""

        widget_manager = self.task.widgets()  # type: WidgetsManager

        self.assertEqual(len(widget_manager), 1)

        widget_manager.delete_all_widgets()

        self.assertEqual(len(widget_manager), 0)
        self.assertEqual(len(self.task.widgets()), 0)

    def test_delete_widget(self):
        """Delete single widget from an activity"""

        widget_manager = self.task.widgets()  # type: WidgetsManager

        self.assertEqual(len(widget_manager), 1)

        widget_manager.delete_widget(0)

        self.assertEqual(len(widget_manager), 0)
        self.assertEqual(len(self.task.widgets()), 0)

    def test_notebook_widget(self):
        widget_manager = self.task.widgets()  # type: WidgetsManager
        notebook = self.project.service(name="Service Gears - Successful")
        widget1 = widget_manager.add_notebook_widget(notebook=notebook,
                                                     title=False)

        widget2 = widget_manager.add_notebook_widget(notebook=notebook,
                                                     title="With custom title")

        widget3 = widget_manager.add_notebook_widget(notebook=notebook.id,
                                                     title="With no padding and custom height",
                                                     customHeight=400,
                                                     noPadding=True)

        self.assertIsInstance(widget1, NotebookWidget)
        self.assertIsInstance(widget2, NotebookWidget)
        self.assertIsInstance(widget3, NotebookWidget)
        self.assertEqual(len(widget_manager), 1 + 3)

        with self.assertRaises(IllegalArgumentError):
            widget_manager.add_notebook_widget(notebook="This will raise an error")

    def test_multicolumn_widget(self):
        widget_manager = self.task.widgets()  # type: WidgetsManager
        bike_part = self.project.part('Bike')
        picture_instance = bike_part.property('Picture')
        picture_model = picture_instance.model()
        multi_column_widget = widget_manager.add_multicolumn_widget(title="Multi column Grid + Attachment")

        widget1 = widget_manager.add_propertygrid_widget(part_instance=bike_part,
                                                         writable_models=[picture_model],
                                                         parent_widget=multi_column_widget)

        widget2 = widget_manager.add_attachmentviewer_widget(
            attachment_property=picture_instance, parent_widget=multi_column_widget.id
        )

        self.assertIsInstance(multi_column_widget, MulticolumnWidget)
        self.assertIsInstance(widget1, PropertygridWidget)
        self.assertIsInstance(widget2, AttachmentviewerWidget)
        self.assertEqual(len(widget_manager), 1 + 3)

    def test_insert_widget(self):
        widget_manager = self.task.widgets()  # type: WidgetsManager
        bike_part = self.project.part('Bike')
        w0 = widget_manager[0]  # meta panel
        w1 = widget_manager.add_propertygrid_widget(part_instance=bike_part,
                                                    writable_models=[bike_part.model().properties],
                                                    title="Original widget 1 (w1)")
        w2 = widget_manager.add_propertygrid_widget(part_instance=bike_part,
                                                    writable_models=[bike_part.model().properties],
                                                    title="Original widget 2 (w2)")
        w3 = widget_manager.add_propertygrid_widget(part_instance=bike_part,
                                                    writable_models=[bike_part.model().properties],
                                                    title="Original widget 3 (w3)")

        # if widget order is `[w0,w1,w2]` and inserting `w3` at index 1 (before Widget1);
        #           index:     0 ^1  2
        # the list will be `[w0,w3,w1,w2]`
        widget_manager.insert(1, w3)

        self.assertTrue(w0.widget_type, WidgetTypes.METAPANEL)
        self.assertEqual(widget_manager[w0.id].order, 0)
        self.assertEqual(widget_manager[w3.id].order, 1)
        self.assertEqual(widget_manager[w1.id].order, 2)
        self.assertEqual(widget_manager[w2.id].order, 3)

        added_widget = widget_manager.add_propertygrid_widget(part_instance=bike_part,
                                                              writable_models=[bike_part.model().properties],
                                                              title="Widget Finally Positioned Under Metapanel",
                                                              order=1)

        self.assertEqual(added_widget.order, 1)
        self.assertEqual(widget_manager[w3.id].order, 2)
        self.assertEqual(widget_manager[w1.id].order, 3)
        self.assertEqual(widget_manager[w2.id].order, 4)

    def test_compatibility_functions(self):
        """Testing various compatibility function for equavalence to the 'customization' in WIM1/PIM1"""

        widget_manager = self.task.widgets()  # type: WidgetsManager

        self.assertTrue(hasattr(widget_manager, 'add_text_widget'))
        self.assertTrue(hasattr(widget_manager, 'add_super_grid_widget'))
        self.assertTrue(hasattr(widget_manager, 'add_property_grid_widget'))
        self.assertTrue(hasattr(widget_manager, 'add_paginated_grid_widget'))
        self.assertTrue(hasattr(widget_manager, 'add_script_widget'))
        self.assertTrue(hasattr(widget_manager, 'add_attachment_viewer_widget'))
        self.assertTrue(hasattr(widget_manager, 'add_navigation_bar_widget'))


class TestWidgetNavigationBarWidget(TestBetamax):

    def setUp(self):
        super(TestWidgetNavigationBarWidget, self).setUp()
        self.task = self.project.create_activity(name="widget_test_task")  # type: Activity2

        self.frame = self.project.part(name='Frame')
        self.frame_model = self.project.model(name='Frame')

        activity_1 = self.project.activity('Task - Basic Table')
        activity_2 = self.project.activity('Task - Form')
        activity_3 = self.project.activity('Task - Service Increase Gears')

        self.nav_bar_config = [
            {'activityId': activity_1, 'emphasize': True},
            {'activityId': activity_2, 'customText': 'Hit me baby!'},
            {'activityId': activity_3}
        ]

    def tearDown(self):
        self.task.delete()
        super(TestWidgetNavigationBarWidget, self).tearDown()

    def test_add_navbar_widget(self):
        widget_manager = self.task.widgets()

        widget = widget_manager.add_tasknavigationbar_widget(
            activities=self.nav_bar_config,
            title="Navbar",
            alignment=NavigationBarAlignment.LEFT
        )

        self.assertIsInstance(widget, TasknavigationbarWidget)
        self.assertEqual(len(widget_manager), 1 + 1)

    def test_add_navbar_widget_in_place(self):
        """
        Test whether the inputs to the function are updated in-place (incorrect) or not (correct).
        For example, activity objects are replaced by their IDs and customText is converted to a string.
        """

        widget_manager = self.task.widgets()

        self.nav_bar_config[0]['customText'] = False

        import copy
        original_bar = copy.deepcopy(self.nav_bar_config)

        widget_manager.add_tasknavigationbar_widget(
            activities=self.nav_bar_config,
            alignment=NavigationBarAlignment.LEFT,
        )

        self.assertEqual(original_bar, self.nav_bar_config)

    def test_add_navbar_widget_incorrect_keys(self):

        widget_manager = self.task.widgets()

        self.nav_bar_config[0]['emphasizeButton'] = True

        with self.assertRaises(IllegalArgumentError):
            widget_manager.add_tasknavigationbar_widget(
                activities=self.nav_bar_config,
                alignment=NavigationBarAlignment.LEFT,
            )

    def test_add_navbar_external_link(self):
        widget_manager = self.task.widgets()

        link = 'https://www.google.com'

        button_1 = self.nav_bar_config[0]
        button_1.pop('activityId')
        button_1['link'] = link

        widget = widget_manager.add_tasknavigationbar_widget(
            activities=self.nav_bar_config,
            alignment=NavigationBarAlignment.LEFT,
        )

        self.assertEqual(widget.meta['taskButtons'][0]['link'], link)

    def test_add_navbar_disabled_button(self):
        widget_manager = self.task.widgets()

        button_1 = self.nav_bar_config[0]
        button_1['isDisabled'] = True

        widget = widget_manager.add_tasknavigationbar_widget(
            activities=self.nav_bar_config,
            alignment=NavigationBarAlignment.LEFT,
        )

        self.assertTrue(widget.meta['taskButtons'][0]['isDisabled'])


class TestWidgetsCopyMove(TestBetamax):
    def setUp(self):
        super(TestWidgetsCopyMove, self).setUp()
        self.task = self.project.create_activity(name="widget_test_task")  # type: Activity2
        self.task_2 = self.project.create_activity(name="test_copy_widget")  # type: Activity2

    def tearDown(self):
        self.task.delete()
        self.task_2.delete()
        super(TestWidgetsCopyMove, self).tearDown()

    def test_copy_widget(self):
        # setUp
        title = "Widget to copy"
        bike_part = self.project.part('Bike')
        widget_manager = self.task.widgets()  # type: WidgetsManager
        widget_1 = widget_manager.add_propertygrid_widget(part_instance=bike_part, all_writable=True,
                                                          title=title)
        widget_1.copy(target_activity=self.task_2, order=0)
        widget_manager_2 = self.task_2.widgets()  # type: WidgetsManager
        associated_model = widget_manager_2[0].parts(category=Category.MODEL)[0]

        # testing
        self.assertEqual(widget_manager_2[0].widget_type, WidgetTypes.PROPERTYGRID)
        self.assertEqual(widget_manager_2[0].title, title)
        self.assertTrue(all(prop.output for prop in associated_model.properties))
        self.assertEqual(len(self.task.widgets()), 2)

    def test_copy_widget_with_wrong_target(self):
        # setUp
        title = "Widget to copy"
        bike_part = self.project.part('Bike')
        widget_manager = self.task.widgets()  # type: WidgetsManager
        widget_1 = widget_manager.add_propertygrid_widget(part_instance=bike_part, all_writable=True,
                                                          title=title)

        # testing
        with self.assertRaises(IllegalArgumentError):
            widget_1.copy(target_activity=self.project, order=0)

    def test_move_widget(self):
        # setUp
        title = "Widget to move"
        bike_part = self.project.part('Bike')
        widget_manager = self.task.widgets()  # type: WidgetsManager
        widget_1 = widget_manager.add_propertygrid_widget(part_instance=bike_part, all_writable=True,
                                                          title=title)
        widget_1.move(target_activity=self.task_2, order=0)
        widget_manager_2 = self.task_2.widgets()  # type: WidgetsManager
        associated_model = widget_manager_2[0].parts(category=Category.MODEL)[0]

        # testing
        self.assertEqual(widget_manager_2[0].widget_type, WidgetTypes.PROPERTYGRID)
        self.assertEqual(widget_manager_2[0].title, title)
        self.assertTrue(all(prop.output for prop in associated_model.properties))
        self.assertEqual(len(self.task.widgets()), 1)



import requests
from typing import Optional, Union, List, Text, Dict

from pykechain.enums import TeamRoles
from pykechain.exceptions import IllegalArgumentError, APIError
from pykechain.models.user import User
from .base import Base


class Team(Base):
    """A virtual object representing a KE-chain Team.

    :ivar name: team name
    :ivar id: teamid of the user
    """

    def __init__(self, json, **kwargs):
        """Construct a user from provided json data."""
        super(Team, self).__init__(json, **kwargs)
        self.ref = json.get('ref')

    def _update(self, resource, update_dict=None, params=None, **kwargs):
        """Update the object."""
        url = self._client._build_url(resource, **kwargs)
        response = self._client._request('PUT', url, json=update_dict, params=params)

        if response.status_code != requests.codes.ok:  # pragma: no cover
            raise APIError("Could not update {} ({})".format(self.__class__.__name__, response.json().get('results')))

        self.refresh(json=response.json().get('results')[0])

    def members(self, role=None):
        # type: (Optional[Union[TeamRoles, Text]]) -> List[Dict]
        """Members of the team.

        You may provide the role in the team, to retrieve only the teammmber with that role. Normally there is a
        single owner, that has administration rights of the team. Normal team members do not have any rights to
        administer the team itself such as altering the teamname, team image and team members. Administrators do
        have the right to administer the the team members.

        :param role: (optional) member belonging to a role :class:`pykechain.enums.TeamRoles` to return.
        :type role: basestring or None
        :raises IllegalArgumentError: when providing incorrect roles
        :return: list of dictionaries with members (pk, username, role, email)

        Example
        -------
        >>> my_team = client.team(name='My own team')
        >>> my_team.members()
        [{"pk":1, "username"="first user", "role"="OWNER", "email":"email@address.com"}, ...]

        """
        if role and role not in TeamRoles.values():
            raise IllegalArgumentError("role should be one of `TeamRoles` {}, got '{}'".format(TeamRoles.values(),
                                                                                               role))

        member_list = self._json_data.get('members')
        if role:
            return [teammember for teammember in member_list if teammember.get('role') == role]
        else:
            return member_list

    def add_members(self, users=None, role=TeamRoles.MEMBER):
        # type: (Optional[List[Union[User, Text]]], Optional[Union[TeamRoles, Text]]) -> ()
        """Members to add to a team.

        :param users: list of members, either `User` objects or usernames
        :type users: List of `User` or List of pk
        :param role: (optional) role of the users to add (default `TeamRoles.MEMBER`)
        :type role: basestring
        :raises IllegalArgumentError: when providing incorrect roles

        Example
        -------
        >>> my_team = client.team(name='My own team')
        >>> other_user = client.users(name='That other person')
        >>> myself = client.users(name='myself')
        >>> my_team.add_members([myself], role=TeamRoles.MANAGER)
        >>> my_team.add_members([other_user], role=TeamRoles.MEMBER)

        """
        if role and role not in TeamRoles.values():
            raise IllegalArgumentError("role should be one of `TeamRoles` {}, got '{}'".format(TeamRoles.values(),
                                                                                               role))

        if not users or not isinstance(users, (list, tuple, set)):
            raise IllegalArgumentError("users should be a list of user_ids or `User` objects, got '{}'".
                                       format(users))

        update_dict = dict(role=role)

        if all(isinstance(user, int) for user in users):
            update_dict['users'] = users
        elif all(isinstance(user, User) for user in users):
            update_dict['users'] = [user.id for user in users]
        else:
            raise IllegalArgumentError("All users should be a list of user_ids or `User` objects, got '{}'".
                                       format(users))

        self._update('team_add_members', team_id=self.id, update_dict=update_dict)

    def remove_members(self, users=None):
        # type: (Optional[List[Union[User, Text]]]) -> ()
        """Members to add to a team.

        :param users: list of members, either `User` objects or usernames
        :type users: List of `User` or List of pk
        :raises IllegalArgumentError: when providing incorrect roles


        Example
        -------
        >>> my_team = client.team(name='My own team')
        >>> other_user = client.users(name='That other person')
        >>> my_team.remove_members([other_user])

        """
        if not users or not isinstance(users, (list, tuple, set)):
            raise IllegalArgumentError("Member should be a list of user_ids or `User` objects, got '{}'".
                                       format(users))

        update_dict = dict()

        if all(isinstance(user, int) for user in users):
            update_dict['users'] = users
        elif all(isinstance(user, User) for user in users):
            update_dict['users'] = [user.id for user in users]
        else:
            raise IllegalArgumentError("All users should be a list of user_ids or `User` objects, got '{}'".
                                       format(users))

        self._update('team_remove_members',
                     update_dict=update_dict,
                     team_id=self.id)

    def scopes(self, **kwargs):
        """Scopes associated to the team."""
        return self._client.scopes(team=self.id, **kwargs)



from typing import List, Optional, Text, Union, Any

from six import text_type, string_types

from pykechain.enums import Category, FilterType, PropertyType
from pykechain.exceptions import IllegalArgumentError
from pykechain.models.part2 import Part2
from pykechain.models.property2 import Property2
from pykechain.utils import is_uuid


class MultiReferenceProperty2(Property2):
    """A virtual object representing a KE-chain multi-references property.

    .. versionadded:: 1.14
    """

    def __init__(self, json, **kwargs):
        """Construct a MultiReferenceProperty from a json object."""
        super(MultiReferenceProperty2, self).__init__(json, **kwargs)

        self._cached_values = None

    @property
    def value(self):
        """Value of a reference property.

        You can set the reference with a Part, Part id or None value.
        Ensure that the model of the provided part, matches the configured model

        :return: a :class:`Part` or None
        :raises APIError: When unable to find the associated :class:`Part`

        Example
        -------
        Get the wheel reference property

        >>> part = project.part('Bike')
        >>> wheels_ref_property = part.property('Wheels')
        >>> isinstance(wheels_ref_property, MultiReferenceProperty2)
        True

        The value returns a list of Parts or is an empty list

        >>> type(wheels_ref_property.value) in (list, tuple)
        True

        Get the selection of wheel instances:

        >>> wheel_choices = wheels_ref_property.choices()

        Choose random wheel from the wheel_choices:

        >>> from random import choice
        >>> wheel_choice_1 = choice(wheel_choices)
        >>> wheel_choice_2 = choice(wheel_choices)

        Set chosen wheel
        1: provide a single wheel:

        >>> wheels_ref_property.value = [wheel_choice_1]

        2: provide multiple wheels:

        >>> wheels_ref_property.value = [wheel_choice_1, wheel_choice_2]

        """
        if not self._value:
            return None
        if not self._cached_values and isinstance(self._value, (list, tuple)):
            assert all([isinstance(v, dict) for v in self._value]), \
                "Expect all elements in the _value to be a dict with 'name' and 'id', got '{}'".format(self._value)
            ids = [v.get('id') for v in self._value]
            # TODO: this will change when the configured model is put in the value_options. Now we configure a single
            # ...   part_model_id in the multirefproperty_model.value as
            if self.category == Category.MODEL and len(ids) == 1:
                self._cached_values = [self._client.part(pk=ids[0], category=None)]
            elif self.category == Category.INSTANCE:
                referred_partmodels = self.model().value
                if referred_partmodels:
                    referred_partmodel = referred_partmodels[0]  # get the only one, which is the right part_model
                else:
                    referred_partmodel = None
                self._cached_values = list(self._client.parts(id__in=','.join(ids),
                                                              model=referred_partmodel,
                                                              category=None))
        return self._cached_values

    @value.setter
    def value(self, value):
        # the dirty 'value' is checked and sanitised an put into value_to_set
        value_to_set = []
        if isinstance(value, (list, tuple)):
            for item in value:
                if isinstance(item, Part2):
                    value_to_set.append(item.id)
                elif isinstance(item, (string_types, text_type)) and is_uuid(item):
                    # tested against a six.text_type (found in the requests' urllib3 package) for unicode
                    # conversion in py27
                    value_to_set.append(item)
                else:
                    raise ValueError("Reference must be a Part, Part id or None. type: {}".format(type(item)))
        elif isinstance(value, type(None)):
            # clear out the list
            value_to_set = None
        else:
            raise ValueError(
                "Reference must be a list (or tuple) of Part, Part id or None. type: {}".format(type(value)))

        # consistency check for references model that should include a scope_id in the value_options.
        if self._options and 'scope_id' not in self._options and value_to_set is not None:
            # if value_to_set is not None, retrieve the scope_id from the first value_to_set
            # we do this smart by checking if we got provided a full Part; that is easier.
            if isinstance(value[0], Part2):
                x_scope_id = value[0].scope_id
            else:
                # retrieve the scope_id from the model value[0] (which is a part in a scope (x_scope))
                referenced_model = self.model().value
                if not referenced_model or not isinstance(referenced_model[0], Part2):
                    # if the referenced model is not set or the referenced value is not a Part set to current scope
                    x_scope_id = self.scope_id
                else:
                    # get the scope_id from the referenced model
                    x_scope_id = referenced_model[0].scope_id
            self._options['scope_id'] = x_scope_id

            # edit the model of the property, such that all instances are updated as well.
            self.model().edit(options=self._options)

        # do the update
        self._put_value(value_to_set)

    def choices(self):
        """Retrieve the parts that you can reference for this `MultiReferenceProperty`.

        This method makes 2 API calls: 1) to retrieve the referenced model, and 2) to retrieve the instances of
        that model.

        :return: the :class:`Part`'s that can be referenced as a :class:`~pykechain.model.PartSet`.
        :raises APIError: When unable to load and provide the choices

        Example
        -------
        >>> property = project.part('Bike').property('a_multi_reference_property')
        >>> reference_part_choices = property.choices()

        """
        # from the reference property (instance) we need to get the value of the reference property in the model
        # in the reference property of the model the value is set to the ID of the model from which we can choose parts
        # the configuration of this ref prop is stored in the model() of this multi-ref-prop. Need to extract model_id
        choices_model_id = self.model()._value[0].get('id')
        possible_choices = self._client.parts(model_id=choices_model_id)  # makes multiple parts call

        return possible_choices

    def set_prefilters(self, property_models, values, filters_type=FilterType.CONTAINS, overwrite=False):
        # type: (List[Union[Text, Part2]], List[Any], List[FilterType], Optional[bool]) -> None
        """
        Set the pre-filters on a `MultiReferenceProperty`.

        :param property_models: `list` of `Property` models (or their IDs) to set pre-filters on
        :type property_models: list
        :param values: `list` of values to pre-filter on, value has to match the property type.
        :type values: list
        :param filters_type: `list` of filter types per pre-fitler, one of :class:`enums.FilterType`,
                defaults to `FilterType.CONTAINS`
        :type filters_type: list
        :param overwrite: determines whether the pre-filters should be over-written or not, defaults to False
        :type overwrite: bool

        :raises IllegalArgumentError: when the type of the input is provided incorrect.
        """
        if any(len(lst) != len(property_models) for lst in [values, filters_type]):
            raise IllegalArgumentError('The lists of "property_models", "values" and "filters_type" should be the '
                                       'same length')
        if not overwrite:
            initial_prefilters = self._options.get('prefilters', {})
            list_of_prefilters = initial_prefilters.get('property_value', [])
            if list_of_prefilters:
                list_of_prefilters = list_of_prefilters.split(',')
        else:
            list_of_prefilters = list()

        for property_model in property_models:
            index = property_models.index(property_model)
            if not isinstance(property_model, Property2):
                if is_uuid(property_model):
                    property_model = self._client.property(id=property_model, category=Category.MODEL)
                else:
                    raise IllegalArgumentError('Pre-filters can only be set on `Property` models or their UUIDs, '
                                               'found type "{}"'.format(type(property_model)))

            if property_model.category != Category.MODEL:
                raise IllegalArgumentError('Pre-filters can only be set on property models, found category "{}"'
                                           'on property "{}"'.format(property_model.category, property_model.name))
            # TODO when a property is freshly created, the property has no "part_id" key in json_data.
            elif self.value[0].id != property_model._json_data.get('part_id', self.value[0].id):
                raise IllegalArgumentError(
                    "Pre-filters can only be set on properties belonging to the referenced Part model, found "
                    "referenced Part model '{}' and Properties belonging to '{}'".format(self.value[0].name,
                                                                                         property_model.part.name))
            else:
                property_id = property_model.id
                if property_model.type == PropertyType.DATETIME_VALUE:
                    datetime_value = values[index]
                    # TODO really not elegant way to deal with DATETIME. Issue is that the value is being double
                    #  encoded (KEC-20504)
                    datetime_value = "{}-{}-{}T00%253A00%253A00.000Z".format(datetime_value.year,
                                                                             str(datetime_value.month).rjust(2, '0'),
                                                                             str(datetime_value.day).rjust(2, '0'))
                    new_prefilter_list = [property_id, datetime_value, filters_type[index]]
                    new_prefilter = '%3A'.join(new_prefilter_list)
                else:
                    if property_model.type == PropertyType.BOOLEAN_VALUE:
                        values[index] = str(values[index]).lower()
                    new_prefilter_list = [property_id, str(values[index]), filters_type[index]]
                    new_prefilter = ':'.join(new_prefilter_list)
                list_of_prefilters.append(new_prefilter)

        options_to_set = self._options
        options_to_set['prefilters'] = {'property_value': ','.join(list_of_prefilters) if list_of_prefilters else {}}

        self.edit(options=options_to_set)

    def set_excluded_propmodels(self, property_models, overwrite=False):
        # type: (List[Union[Text, Part2]], Optional[bool]) -> None
        """
        Exclude a list of properties from being visible in the part-shop and modal (pop-up) of the reference property.

        :param property_models: `list` of Property2 models (or their IDs) to exclude.
        :type property_models: list
        :param overwrite: flag whether to overwrite existing (True) or append (False, default) to existing filter(s).
        :type overwrite bool

        :raises IllegalArgumentError
        """
        if not overwrite:
            list_of_propmodels_excl = self._options.get('propmodels_excl', [])
        else:
            list_of_propmodels_excl = list()

        for property_model in property_models:
            if not isinstance(property_model, Property2):
                if is_uuid(property_model):
                    property_model = self._client.property(id=property_model, category=Category.MODEL)
                else:
                    raise IllegalArgumentError('A part reference property can only exclude `Property` models or their '
                                               'UUIDs, found type "{}"'.format(type(property_model)))
            if property_model.category != Category.MODEL:
                raise IllegalArgumentError("A part reference property can only excluded `Property` models, found "
                                           "category '{}' on property '{}'".format(property_model.category,
                                                                                   property_model.name))
            # TODO when a property is freshly created, the property has no "part_id" key in json_data.
            elif self.value[0].id != property_model._json_data.get('part_id', self.value[0].id):
                raise IllegalArgumentError(
                    'A part reference property can only exclude properties belonging to the referenced Part model, '
                    'found referenced Part model "{}" and Properties belonging to "{}"'.format(
                        self.value[0].name, property_model.part.name))
            else:
                list_of_propmodels_excl.append(property_model.id)

        options_to_set = self._options
        options_to_set['propmodels_excl'] = list_of_propmodels_excl

        self.edit(options=options_to_set)



import datetime
from time import sleep
from unittest import skipIf

from pykechain.enums import ScopeStatus
from pykechain.exceptions import NotFoundError, MultipleFoundError, IllegalArgumentError
from pykechain.models import Team
from tests.classes import TestBetamax
from tests.utils import TEST_FLAG_IS_PIM3


class TestScopes(TestBetamax):

    def test_retrieve_scopes(self):
        self.assertTrue(self.client.scopes())

    def test_scope_attributes(self):
        attributes = ['_client', '_json_data', 'id', 'name', 'created_at', 'updated_at', 'ref',
                      'process', 'workflow_root', 'description', 'status', 'category', 'tags',
                      'start_date', 'due_date']

        obj = self.project
        for attribute in attributes:
            self.assertTrue(hasattr(obj, attribute),
                            "Could not find '{}' in the object: '{}'".format(attribute, obj.__dict__.keys()))

    def test_retrieve_scope_with_kwargs(self):
        retrieve_scopes_with_kwargs = self.client.scopes(name__icontains='Bike')
        self.assertTrue(retrieve_scopes_with_kwargs)
        self.assertTrue(len(retrieve_scopes_with_kwargs) >= 1)

    def test_retrieve_single_unknown(self):
        with self.assertRaises(NotFoundError):
            self.client.scope('123lladadwd')

    def test_retrieve_single_multiple(self):
        with self.assertRaises(MultipleFoundError):
            self.client.scope()

    def test_add_member(self):
        member_to_be_added = 'anotheruser'
        # testing
        self.project.add_member(member_to_be_added)
        self.project = self.client.scope(pk=self.project.id)
        project_members = self.project.members()
        self.assertTrue(member_to_be_added in [member['username'] for member in project_members])
        # teardown
        self.project.remove_member(member_to_be_added)

    def test_add_member_by_id(self):
        member_id = 4
        with self.assertRaises(TypeError):
            self.project.add_member(member_id)

    def test_add_non_existing_member(self):
        member_to_be_added = 'Nobody'
        with self.assertRaises(NotFoundError):
            self.project.add_member(member_to_be_added)

    def test_remove_member(self):
        member_to_be_removed = 'anotheruser'
        # setUp
        self.project.add_member(member_to_be_removed)
        # testing
        self.project.remove_member(member_to_be_removed)
        self.project = self.client.scope(pk=self.project.id)
        project_members = self.project.members()
        self.assertTrue(member_to_be_removed not in [member['username'] for member in project_members])

    def test_add_manager(self):
        manager_to_be_added = 'anotheruser'
        # testing
        self.project.add_manager(manager_to_be_added)
        self.project = self.client.scope(pk=self.project.id)
        project_managers = self.project.members(is_manager=True)
        self.assertTrue(manager_to_be_added in [manager['username'] for manager in project_managers])
        # teardown
        self.project.remove_member(manager_to_be_added)

    def test_remove_manager(self):
        manager_to_be_removed = 'anotheruser'
        # setUp
        self.project.add_manager(manager_to_be_removed)
        # testing
        self.project.remove_manager(manager_to_be_removed)
        self.project = self.client.scope(pk=self.project.id)
        project_managers = self.project.members(is_manager=True)
        self.assertTrue(manager_to_be_removed not in [manager['username'] for manager in project_managers])
        project_managers = self.project.members(is_manager=False)
        self.assertTrue(manager_to_be_removed in [manager['username'] for manager in project_managers])
        # teardown
        self.project.remove_member(manager_to_be_removed)

    def test_add_leadmember(self):
        leadmember_to_be_added = 'anotheruser'
        # testing
        self.project.add_leadmember(leadmember_to_be_added)
        self.project = self.client.scope(pk=self.project.id)
        project_leadmembers = self.project.members(is_leadmember=True)
        self.assertTrue(leadmember_to_be_added in [leadmember['username'] for leadmember in project_leadmembers])
        # teardown
        self.project.remove_member(leadmember_to_be_added)

    def test_remove_leadmember(self):
        leadmember_to_be_removed = 'anotheruser'
        # setUp
        self.project.add_leadmember(leadmember_to_be_removed)
        # testing
        self.project.remove_leadmember(leadmember_to_be_removed)
        self.project = self.client.scope(pk=self.project.id)
        project_leadmembers = self.project.members(is_leadmember=True)
        self.assertTrue(leadmember_to_be_removed not in [leadmember['username'] for leadmember in project_leadmembers])
        project_leadmembers = self.project.members(is_leadmember=False)
        self.assertTrue(leadmember_to_be_removed in [leadmember['username'] for leadmember in project_leadmembers])
        # teardown
        self.project.remove_member(leadmember_to_be_removed)

    # 2.2.0+
    def test_edit_scope(self):
        # setUp
        new_scope_name = 'Pykechain testing (bike project)'
        old_scope_name = self.project.name
        new_scope_description = 'Project used to build a Bike. Part-time job is to also test Pykechain.'
        old_scope_description = 'Project used to test Pykechain and push it to the limits'
        new_start_date = datetime.datetime(2018, 12, 5, tzinfo=None)
        old_start_date = datetime.datetime(2017, 4, 12)
        new_due_date = datetime.datetime(2018, 12, 8, tzinfo=None)
        old_due_date = datetime.datetime(2018, 4, 1)
        old_status = ScopeStatus.ACTIVE

        self.project.edit(name=new_scope_name, description=new_scope_description, start_date=new_start_date,
                          due_date=new_due_date, status=ScopeStatus.CLOSED)

        # testing
        retrieved_project = self.client.scope(id=self.project.id, status=ScopeStatus.CLOSED)

        self.assertTrue(retrieved_project.name == new_scope_name)
        self.assertTrue(retrieved_project._json_data['text'] == new_scope_description)
        self.assertTrue(retrieved_project._json_data['start_date'] in ('2018-12-05T00:00:00Z', '2018-12-05T00:00:00+00:00'))
        self.assertTrue(retrieved_project._json_data['due_date'] in ('2018-12-08T00:00:00Z', '2018-12-08T00:00:00+00:00'))
        self.assertTrue(retrieved_project._json_data['status'] == ScopeStatus.CLOSED)

        with self.assertRaises(IllegalArgumentError):
            self.project.edit(name=True)

        with self.assertRaises(IllegalArgumentError):
            self.project.edit(description=new_start_date)

        with self.assertRaises(IllegalArgumentError):
            self.project.edit(start_date='Yes, man!')

        with self.assertRaises(IllegalArgumentError):
            self.project.edit(due_date='Wrong')

        with self.assertRaises(IllegalArgumentError):
            self.project.edit(status='ILLEGAL_STATUS_WILL_CAUSE_ERR')

        # tearDown
        self.project.edit(name=old_scope_name, description=old_scope_description, start_date=old_start_date,
                          due_date=old_due_date, status=old_status)

    # v2.3.3
    def test_edit_scope_team(self):
        """Test capabilities of changing team scope."""

        # setup
        team_url = self.project._client._build_url('teams')

        r = self.project._client._request('get', team_url, params=dict(name='Team No.1'))
        team_kew_id = r.json().get('results')[0].get('id')  # no iffer as it should fail hard when not a respcode =ok

        r = self.project._client._request('get', team_url, params=dict(name='Second Team'))
        team_b_id = r.json().get('results')[0].get('id')

        # save current team
        if self.client.match_app_version(label="pim", version=">=3.0.0"):
            team_dict = self.project._json_data.get('team_id_name')
            saved_team_id = team_dict and team_dict.get('id')
        else:
            team_dict = self.project._json_data.get('team')
            saved_team_id = team_dict and team_dict.get('id')

        self.project.edit(team=team_b_id)
        if self.client.match_app_version(label="pim", version=">=3.0.0"):
            self.assertEqual(self.project._json_data.get('team_id_name').get('id'), team_b_id)
        else:
            self.assertEqual(self.project._json_data.get('team').get('id'), team_b_id)

        self.project.edit(team=team_kew_id)
        if self.client.match_app_version(label="pim", version=">=3.0.0"):
            self.assertEqual(self.project._json_data.get('team_id_name').get('id'), team_kew_id)
        else:
            self.assertEqual(self.project._json_data.get('team').get('id'), team_kew_id)

        self.project.edit(team=saved_team_id)
        if self.client.match_app_version(label="pim", version=">=3.0.0"):
            self.assertEqual(self.project._json_data.get('team_id_name').get('id'), saved_team_id)
        else:
            self.assertEqual(self.project._json_data.get('team').get('id'), saved_team_id)

    def test_team_property_of_scope(self):
        """Test for the property 'team' of a scope."""
        team = self.project.team
        self.assertIsInstance(team, Team)


@skipIf(not TEST_FLAG_IS_PIM3, reason="This tests is designed for PIM version 2, expected to fail on old PIM")
class TestScopes2SpecificTests(TestBetamax):

    def setUp(self):
        """
        SetUp the test.

        When you create a scope, assign it to self.scope such that it is deleted when the test is complete

        :ivar scope: scope that you create during a test, which will be autoremoved in the cleanup.
        """
        super(TestScopes2SpecificTests, self).setUp()
        self.scope = None

    def tearDown(self):
        """Teardown the test, remove the scope.

        When you create a scope, assign it to self.scope such that it is deleted when the test is complete
        """
        if self.scope:
            self.scope.delete()
        super(TestScopes2SpecificTests, self).tearDown()

    def test_retrieve_scope2_members(self):
        original_scope_members = [u.get('username') for u in self.project.members()]
        original_scope_managers = [u.get('username') for u in self.project.members(is_manager=True)]
        scope_managers = original_scope_managers + ['testmanager']
        scope_members = original_scope_members + ['anotheruser', 'testuser']

        self.project.add_manager('testmanager')
        self.project.add_member('anotheruser')
        self.project.add_member('testuser')

        self.project.refresh()

        members_usernames = [member['username'] for member in self.project.members()]
        managers_usernames = [manager['username'] for manager in self.project.members(is_manager=True)]

        self.assertSetEqual(set(scope_members), set(members_usernames))
        self.assertSetEqual(set(scope_managers), set(managers_usernames))

        # teardown
        self.project.remove_member('anotheruser')

    def test_clone_scope_vanilla(self):
        self.scope = self.project.clone(asynchronous=False)

        self.assertNotEqual(self.project.id, self.scope.id)
        self.assertEqual(self.project._json_data.get('tasks_count'), self.scope._json_data.get('tasks_count'))
        self.assertEqual(self.project._json_data.get('processes_count'), self.scope._json_data.get('processes_count'))
        self.assertEqual(self.project._json_data.get('activities_count'), self.scope._json_data.get('activities_count'))
        self.assertEqual(self.project._json_data.get('parts_count'), self.scope._json_data.get('parts_count'))
        self.assertEqual(self.project._json_data.get('model_parts_count'), self.scope._json_data.get('model_parts_count'))
        self.assertEqual(self.project._json_data.get('properties_count'), self.scope._json_data.get('properties_count'))

    def test_scope_delete(self):
        new_scope = self.project.clone(asynchronous=False)
        self.assertNotEqual(self.project.id, new_scope.id)
        new_scope.delete(asynchronous=False)

        with self.assertRaisesRegex(NotFoundError, 'No scope fits criteria'):
            # throw in arbitrary sleep to give backend time to actually delete the scope.
            sleep(1)
            self.client.scope(pk=new_scope.id)

    def test_clone_scope_updated_name_description_tags_etc(self):
        """
        self, source_scope, name=None, status=None, tags=None, start_date=None, due_date=None,
                    description=None, team=None, scope_options=None, async=False):
        """
        new_scope = self.project.clone(
            name="new bike scope",
            description="new description",
            status=ScopeStatus.CLOSED,
            # scope_options={"hello": "world"},
            tags=('a', 'b', 'b', 'c'),
            asynchronous=False)
        self.scope = new_scope

        self.assertNotEqual(self.project.id, new_scope.id)
        self.assertEqual(new_scope._json_data.get('name'), "new bike scope")
        self.assertEqual(new_scope._json_data.get('text'), "new description")
        self.assertEqual(new_scope._json_data.get('status'), ScopeStatus.CLOSED)
        # self.assertListEqual(new_scope._json_data.get('tags'), ["a", "b", "c"])



from pykechain.enums import Category, Multiplicity
from pykechain.exceptions import NotFoundError, MultipleFoundError
from pykechain.models import PartSet
from tests.classes import TestBetamax


class TestPartRetrieve(TestBetamax):
    # 1.8
    def test_get_instances_of_a_model(self):
        wheel_model = self.project.model('Wheel')
        wheel_instances = wheel_model.instances()

        self.assertIsInstance(wheel_instances, PartSet)
        for wheel_instance in wheel_instances:
            self.assertEqual(wheel_instance.category, Category.INSTANCE)
            self.assertEqual(wheel_instance.model().id, wheel_model.id)

    def test_get_instances_of_an_instances_raises_notfound(self):
        wheel_instance = self.project.part('Rear Wheel', category=Category.INSTANCE)
        with self.assertRaises(NotFoundError):
            wheel_instance.instances()

    def test_get_single_instance_of_a_model(self):
        bike_model = self.project.model('Bike')
        bike_instance = bike_model.instance()

        self.assertEqual(bike_instance.category, Category.INSTANCE)

    def test_get_single_instance_of_a_multiplicity_model_raises_multiplefounderror(self):
        wheel_model = self.project.model('Wheel')

        with self.assertRaises(MultipleFoundError):
            wheel_model.instance()

    # test added in 1.12.7
    def test_get_single_instance_of_a_model_without_instances_raises_notfounderror(self):
        catalog = self.project.model(name__startswith='Catalog')
        model_without_instances = self.project.create_model(parent=catalog, name='model_without_instances',
                                                            multiplicity=Multiplicity.ZERO_ONE)

        with self.assertRaises(NotFoundError):
            model_without_instances.instance()

        # tearDown
        model_without_instances.delete()

    # test added in 2.1
    def test_get_parts_with_descendants_tree(self):
        # setUp
        bike_part = self.project.part(name='Bike')
        bike_part.populate_descendants()
        bike_model = self.project.model(name='Bike')
        bike_model.populate_descendants()
        # testing
        self.assertTrue(len(bike_part._cached_children) >= 5)
        self.assertTrue(len(bike_model._cached_children) >= 4)





import datetime
import os
import time
import warnings
from collections import Iterable
from typing import Any, List, Union, Text

import requests
from requests.compat import urljoin  # type: ignore
from six import text_type

from pykechain.defaults import ASYNC_REFRESH_INTERVAL, ASYNC_TIMEOUT_LIMIT, API_EXTRA_PARAMS
from pykechain.enums import ActivityType, ActivityStatus, Category, ActivityClassification, ActivityRootNames, \
    PaperSize, PaperOrientation
from pykechain.exceptions import NotFoundError, IllegalArgumentError, APIError
from pykechain.models import Base
from pykechain.models.tags import TagsMixin
from pykechain.models.widgets.widgets_manager import WidgetsManager
from pykechain.utils import is_uuid, parse_datetime


class Activity2(Base, TagsMixin):
    """A virtual object representing a KE-chain activity.

    .. versionadded:: 2.0

    :ivar id: id of the activity
    :type id: uuid
    :ivar name: name of the activity
    :type name: basestring
    :ivar created_at: created datetime of the activity
    :type created_at: datetime
    :ivar updated_at: updated datetime of the activity
    :type updated_at: datetime
    :ivar description: description of the activity
    :type description: basestring
    :ivar status: status of the activity. One of :class:`pykechain.enums.ActivityStatus`
    :type status: basestring
    :ivar classification: classification of the activity. One of :class:`pykechain.enums.ActivityClassificiation`
    :type classification: basestring
    :ivar activity_type: Type of the activity. One of :class:`pykechain.enums.ActivityType` for WIM version 2
    :type activity_type: basestring
    """

    def __init__(self, json, **kwargs):
        """Construct an Activity from a json object."""
        super(Activity2, self).__init__(json, **kwargs)

        self._scope_id = json.get('scope_id')

        self.ref = json.get('ref')
        self.description = json.get('description', '')
        self.status = json.get('status')
        self.classification = json.get('classification')
        self.activity_type = json.get('activity_type')
        self.start_date = parse_datetime(json.get('start_date'))
        self.due_date = parse_datetime(json.get('due_date'))

        self._tags = json.get('tags')

    def refresh(self, *args, **kwargs):
        """Refresh the object in place."""
        super(Activity2, self).refresh(url=self._client._build_url('activity', activity_id=self.id),
                                       extra_params=API_EXTRA_PARAMS['activity'])

    #
    # additional properties
    #

    @property
    def assignees(self):
        """List of assignees to the activity.

        Provides a list of `User` objects or an empty list. If no `assignees_ids` are in the API call, then
        returns None.

        :return: a list of `User`'s or an empty list.
        :rtype: list or None
        """
        if 'assignees_ids' in self._json_data and self._json_data.get('assignees_ids') == list():
            return []
        elif 'assignees_ids' in self._json_data and self._json_data.get('assignees_ids'):
            assignees_ids_str = ','.join([str(id) for id in self._json_data.get('assignees_ids')])
            return self._client.users(id__in=assignees_ids_str, is_hidden=False)
        return None

    @property
    def scope_id(self):
        """
        Id of the scope this Activity belongs to.

        This property will always produce a scope_id, even when the scope object was not included in an earlier
        response.

        When the :class:`Scope` is not included in this task, it will make an additional call to the KE-chain API.

        :return: the scope id
        :type: uuid
        :raises NotFoundError: if the scope could not be found
        """
        if self._scope_id is None:
            self.refresh()
            if self._scope_id is None:
                raise NotFoundError("This activity '{}'({}) does not belong to a scope, something is weird!".
                                    format(self.name, self.id))
        return self._scope_id

    @property
    def scope(self):
        """
        Scope this Activity belongs to.

        This property will return a `Scope` object. It will make an additional call to the KE-chain API.

        :return: the scope
        :type: :class:`pykechain.models.Scope`
        :raises NotFoundError: if the scope could not be found
        """
        return self._client.scope(pk=self.scope_id, status=None)

    #
    # predicates
    #

    def is_rootlevel(self):
        """
        Determine if the Activity is at the toplevel of a project.

        It will look for the name of the parent which should be either ActivityRootNames.WORKFLOW_ROOT or
        ActivityRootNames.CATALOG_ROOT. If the name of the parent cannot be found an additional API call is made
        to retrieve the parent object (based on the `parent_id` in the json_data).

        :return: Return True if it is a root level activity, otherwise return False
        :rtype: bool
        """
        # when the activity itself is a root, than return False immediately
        if self.is_root():
            return False

        parent_name = None
        parent_dict = self._json_data.get('parent_id_name')

        if parent_dict and 'name' in parent_dict:
            parent_name = parent_dict.get('name')
        if not parent_dict:
            parent_name = self._client.activity(id=self._json_data.get('parent_id')).name
        if parent_name in ActivityRootNames.values():
            return True
        return False

    def is_task(self):
        """
        Determine if the Activity is of ActivityType.TASK.

        :return: Return True if it is a task, otherwise return False
        :rtype: bool
        """
        return self.activity_type == ActivityType.TASK

    def is_subprocess(self):
        """
        Determine if the Activity is of ActivityType.PROCESS.

        :return: Return True if it is a subprocess, otherwise return False
        :rtype: bool
        """
        return self.is_process()

    def is_process(self):
        """
        Determine if the Activity is of ActivityType.PROCESS.

        :return: Return True if it is a process, otherwise return False
        :rtype: bool
        """
        return self.activity_type == ActivityType.PROCESS

    def is_workflow(self):
        """
        Determine if the classification of the Activity is of ActivityClassification.WORKFLOW.

        :return: Return True if it is a workflow classification activity, otherwise return False
        :rtype: bool
        """
        return self.classification == ActivityClassification.WORKFLOW

    def is_catalog(self):
        """
        Determine if the classification of the Activity is of ActivityClassification.CATALOG.

        :return: Return True if it is a catalog classification activity, otherwise return False
        :rtype: bool
        """
        return self.classification == ActivityClassification.CATALOG

    def is_workflow_root(self):
        """
        Determine if the classification of the Activity is of ActivityClassification.WORKFLOW and a ROOT object.

        :return: Return True if it is a root workflow classification activity, otherwise return False
        :rtype: bool
        """
        return self.is_root() and self.is_workflow()

    def is_catalog_root(self):
        """
        Determine if the classification of the Activity is of ActivityClassification.CATALOG and a ROOT object.

        :return: Return True if it is a root catalog classification activity, otherwise return False
        :rtype: bool
        """
        return self.is_root() and self.is_catalog()

    def is_root(self):
        """
        Determine if the Activity is a ROOT object.

        If you want to determine if it is also a workflow or a catalog root, use :func:`Activity.is_workflow_root()`
        or :func:`Activity.is_catalog_root()` methods.

        :return: Return True if it is a root object, otherwise return False
        :rtype: bool
        """
        return self.name in ActivityRootNames.values() and self._json_data.get('parent_id', False) is None

    def is_configured(self):
        """
        Determine if the Activity is configured with input and output properties.

        Makes an additional lightweight call to the API to determine if any associated models are there.

        :return: Return True if it is configured, otherwise return False
        :rtype: bool
        """
        # check configured based on if we get at least 1 part back
        associated_models = self.parts(category=Category.MODEL, limit=1)
        if associated_models:
            return True
        else:
            return False

    def is_customized(self):
        """
        Determine if the Activity is customized.

        In other words if it has a customization. Use can use the :func:`Activity.customization()` to retrieve
        the customization object and configure the task.

        :return: Return True if it is customized, otherwise return False
        :rtype: bool
        """
        if self._json_data.get('customization', False):
            return True
        else:
            return False

    #
    # methods
    #

    def create(self, *args, **kwargs):
        """Create a new activity belonging to this subprocess.

        See :func:`pykechain.Client.create_activity` for available parameters.

        :raises IllegalArgumentError: if the `Activity2` is not a `PROCESS`.
        :raises APIError: if an Error occurs.
        """
        if self.activity_type != ActivityType.PROCESS:
            raise IllegalArgumentError("One can only create a task under a subprocess.")
        return self._client.create_activity(self.id, *args, **kwargs)

    def subprocess(self):
        """Retrieve the subprocess in which this activity is defined.

        .. warning::
            This method is deprecated for newer releases of KE-chain (version 2.9.0 and higher). Please
            use the :func:`Activity2.parent()` method.

        If this is a task on top level, it raises NotFounderror.

        :return: a subprocess :class:`Activity2`
        :raises NotFoundError: when it is a task in the top level of a project
        :raises APIError: when other error occurs

        """
        raise ('Subprocess function is outdated in KE-chain 2.9.0, use `Activity2.parent()` method', DeprecationWarning)
        # return self.parent()

    def parent(self):
        """Retrieve the parent in which this activity is defined.

        If this is a task on top level, it raises NotFounderror.

        :return: a :class:`Activity2`
        :raises NotFoundError: when it is a task in the top level of a project
        :raises APIError: when other error occurs

        Example
        -------
        >>> task = project.activity('Subtask')
        >>> parent_of_task = task.parent()

        """
        parent_id = self._json_data.get('parent_id')
        if parent_id is None:
            raise NotFoundError("Cannot find subprocess for this task '{}', "
                                "as this task exist on top level.".format(self.name))
        return self._client.activity(pk=parent_id, scope=self.scope_id)

    def children(self, **kwargs):
        """Retrieve the direct activities of this subprocess.

        It returns a combination of Tasks (a.o. UserTasks) and Subprocesses on the direct descending level.
        Only when the activity is a Subprocess, otherwise it raises a NotFoundError

        :param kwargs: Additional search arguments, check :func:`pykechain.Client.activities` for additional info
        :return: a list of :class:`Activity2`
        :raises NotFoundError: when this task is not of type `ActivityType.PROCESS`

        Example
        -------
        >>> parent = project.parent('Subprocess')
        >>> children = subprocess.children()

        Example searching for children of a subprocess which contains a name (icontains searches case insensitive

        >>> parent = project.parent('Subprocess')
        >>> children = subprocess.children(name__icontains='more work')

        """
        if self.activity_type != ActivityType.PROCESS:
            raise NotFoundError("Only subprocesses can have children, please choose a subprocess instead of a '{}' "
                                "(activity '{}')".format(self.activity_type, self.name))

        return self._client.activities(parent_id=self.id, scope=self.scope_id, **kwargs)

    def siblings(self, **kwargs):
        """Retrieve the other activities that also belong to the parent.

        It returns a combination of Tasks (a.o. UserTasks) and Subprocesses on the level of the current task, including
        itself. This also works if the activity is of type `ActivityType.PROCESS`.

        :param kwargs: Additional search arguments, check :func:`pykechain.Client.activities` for additional info
        :return: list of :class:`Activity2`
        :raises NotFoundError: when it is a task in the top level of a project

        Example
        -------
        >>> task = project.activity('Some Task')
        >>> siblings = task.siblings()

        Example for siblings containing certain words in the task name
        >>> task = project.activity('Some Task')
        >>> siblings = task.siblings(name__contains='Another Task')

        """
        parent_id = self._json_data.get('parent_id')
        if parent_id is None:
            raise NotFoundError("Cannot find subprocess for this task '{}', "
                                "as this task exist on top level.".format(self.name))
        return self._client.activities(parent_id=parent_id, scope=self.scope_id, **kwargs)

    def edit(self, name=None, description=None, start_date=None, due_date=None, assignees=None, assignees_ids=None,
             status=None, tags=None):
        # type: (Text, Text, datetime.datetime, datetime.datetime, List[Text], List[Text], Union[ActivityStatus, Text], Iterable[Text]) -> None  # noqa: E501
        """Edit the details of an activity.

        :param name: (optionally) edit the name of the activity
        :type name: basestring or None
        :param description: (optionally) edit the description of the activity
        :type description: basestring or None
        :param start_date: (optionally) edit the start date of the activity as a datetime object (UTC time/timezone
                            aware preferred)
        :type start_date: datetime or None
        :param due_date: (optionally) edit the due_date of the activity as a datetime object (UTC time/timzeone
                            aware preferred)
        :type due_date: datetime or None
        :param assignees: (optionally) edit the assignees (usernames) of the activity as a list, will overwrite all
                          assignees
        :type assignees: list(basestring) or None
        :param assignees_ids: (optionally) edit the assignees (user id's) of the activity as a list, will overwrite all
                             assignees
        :type assignees_ids: list(basestring) or None
        :param status: (optionally) edit the status of the activity as a string based
                       on :class:`~pykechain.enums.ActivityType`
        :type status: basestring or None
        :param tags: (optionally) replace the tags on an activity, which is a list of strings ["one","two","three"]
        :type tags: list of basestring or None

        :raises NotFoundError: if a `username` in the list of assignees is not in the list of scope members
        :raises IllegalArgumentError: if the type of the inputs is not correct
        :raises APIError: if another Error occurs
        :warns: UserWarning - When a naive datetime is provided. Defaults to UTC.

        Example
        -------
        >>> from datetime import datetime
        >>> my_task = project.activity('Specify the wheel diameter')
        >>> my_task.edit(name='Specify wheel diameter and circumference',
        ...              description='The diameter and circumference are specified in inches',
        ...              start_date=datetime.utcnow(),  # naive time is interpreted as UTC time
        ...              assignee='testuser')

        If we want to provide timezone aware datetime objects we can use the 3rd party convenience library :mod:`pytz`.
        Mind that we need to fetch the timezone first and use `<timezone>.localize(<your datetime>)` to make it
        work correctly.

        Using `datetime(2017,6,1,23,59,0 tzinfo=<tz>)` does NOT work for most timezones with a
        daylight saving time. Check the `pytz <http://pythonhosted.org/pytz/#localized-times-and-date-arithmetic>`_
        documentation.

        To make it work using :mod:`pytz` and timezone aware :mod:`datetime` see the following example::

        >>> import pytz
        >>> start_date_tzaware = datetime.now(pytz.utc)
        >>> mytimezone = pytz.timezone('Europe/Amsterdam')
        >>> due_date_tzaware = mytimezone.localize(datetime(2019, 10, 27, 23, 59, 0))
        >>> my_task.edit(due_date=due_date_tzaware, start_date=start_date_tzaware)

        """
        update_dict = {'id': self.id}
        if name is not None:
            if isinstance(name, (str, text_type)):
                update_dict.update({'name': name})
                self.name = name
            else:
                raise IllegalArgumentError('Name should be a string')

        if description is not None:
            if isinstance(description, (str, text_type)):
                update_dict.update({'description': description})
                self.description = description
            else:
                raise IllegalArgumentError('Description should be a string')

        if start_date is not None:
            if isinstance(start_date, datetime.datetime):
                if not start_date.tzinfo:
                    warnings.warn("The startdate '{}' is naive and not timezone aware, use pytz.timezone info. "
                                  "This date is interpreted as UTC time.".format(start_date.isoformat(sep=' ')))
                update_dict.update({'start_date': start_date.isoformat(sep='T')})
            else:
                raise IllegalArgumentError('Start date should be a datetime.datetime() object')

        if due_date is not None:
            if isinstance(due_date, datetime.datetime):
                if not due_date.tzinfo:
                    warnings.warn("The duedate '{}' is naive and not timezone aware, use pytz.timezone info. "
                                  "This date is interpreted as UTC time.".format(due_date.isoformat(sep=' ')))
                update_dict.update({'due_date': due_date.isoformat(sep='T')})
            else:
                raise IllegalArgumentError('Due date should be a datetime.datetime() object')

        if isinstance(assignees_ids, (list, tuple)) or isinstance(assignees, (list, tuple)):
            update_assignees_ids = []
            if isinstance(assignees_ids, (list, tuple)):
                users = self._client.users()
                update_assignees_ids = [u.id for u in users if u.id in assignees_ids]
                if len(update_assignees_ids) != len(assignees_ids):
                    raise NotFoundError("All assignees should be a member of the project")
            elif isinstance(assignees, (list, tuple)):
                users = self._client.users()
                update_assignees_ids = [u.id for u in users if u.username in assignees]
                if len(update_assignees_ids) != len(assignees):
                    raise NotFoundError("All assignees should be a member of the project")
            else:
                raise IllegalArgumentError("Provide the usernames either as list of usernames of user id's")

            if isinstance(update_assignees_ids, list):
                project = self._client.scope(pk=self.scope_id, status=None)
                member_ids_list = [member['id'] for member in project._json_data['members']]
                for assignee_id in update_assignees_ids:
                    if assignee_id not in member_ids_list:
                        raise NotFoundError("Assignee '{}' should be a member of the project".format(assignee_id))
                update_dict.update({'assignees_ids': update_assignees_ids})
        elif assignees_ids or assignees:
            raise IllegalArgumentError("If assignees_ids or assignees are provided, they should be a list or tuple")

        if status is not None:
            if isinstance(status, (str, text_type)) and status in ActivityStatus.values():
                update_dict.update({'status': status})
            else:
                raise IllegalArgumentError('Status should be a string and in the list of acceptable '
                                           'status strings: {}'.format(ActivityStatus.values()))

        if tags is not None:
            if isinstance(tags, (list, tuple, set)):
                update_dict.update({'tags': tags})
            else:
                raise IllegalArgumentError('tags should be a an array (list, tuple, set) of strings')

        url = self._client._build_url('activity', activity_id=self.id)

        response = self._client._request('PUT', url, json=update_dict)

        if response.status_code != requests.codes.ok:  # pragma: no cover
            raise APIError("Could not update Activity ({})".format(response))

        self.refresh(json=response.json().get('results')[0])

    def delete(self):
        """Delete this activity.

        :return: True when successful
        :raises APIError: when unable to delete the activity
        """
        response = self._client._request('DELETE', self._client._build_url('activity', activity_id=self.id))

        if response.status_code != requests.codes.no_content:
            raise APIError("Could not delete activity: {} with id {}".format(self.name, self.id))
        return True

    #
    # Searchers and retrievers
    #

    def parts(self, *args, **kwargs):
        """Retrieve parts belonging to this activity.

        Without any arguments it retrieves the Instances related to this task only.

        This call only returns the configured properties in an activity. So properties that are not configured
        are not in the returned parts.

        See :class:`pykechain.Client.parts` for additional available parameters.

        Example
        -------
        >>> task = project.activity('Specify Wheel Diameter')
        >>> parts = task.parts()

        To retrieve the models only.
        >>> parts = task.parts(category=Category.MODEL)

        """
        return self._client.parts(*args, activity=self.id, **kwargs)

    def associated_parts(self, *args, **kwargs):
        """Retrieve models and instances belonging to this activity.

        This is a convenience method for the :func:`Activity.parts()` method, which is used to retrieve both the
        `Category.MODEL` as well as the `Category.INSTANCE` in a tuple.

        This call only returns the configured properties in an activity. So properties that are not configured
        are not in the returned parts.

        If you want to retrieve only the models associated to this task it is better to use:
            `task.parts(category=Category.MODEL)`.

        See :func:`pykechain.Client.parts` for additional available parameters.

        :returns: a tuple(models of :class:`PartSet`, instances of :class:`PartSet`)

        Example
        -------
        >>> task = project.activity('Specify Wheel Diameter')
        >>> all_models, all_instances = task.associated_parts()

        """
        return (
            self.parts(category=Category.MODEL, *args, **kwargs),
            self.parts(category=Category.INSTANCE, *args, **kwargs)
        )

    #
    # Customizations
    #

    def widgets(self, **kwargs):
        # type: (**Any) -> WidgetsManager
        """
        Widgets of the activity.

        Works with KE-chain version 3.

        :param kwargs: additional keyword arguments
        :return: A :class:`WidgetManager` list, containing the widgets
        :rtype: WidgetManager
        :raises NotFoundError: when the widgets could not be found
        :raises APIError: when the API does not support the widgets, or when the API gives an error.
        """
        return self._client.widgets(activity=self.id, **kwargs)

    def customization(self):
        """
        Get a customization object representing the customization of the activity.

        .. versionadded:: 1.11

        :return: An instance of :class:`customization.ExtCustomization`

        Example
        -------
        >>> activity = project.activity(name='Customizable activity')
        >>> customization = activity.customization()
        >>> part_to_show = project.part(name='Bike')
        >>> customization.add_property_grid_widget(part_to_show,custom_title="My super bike")

        """
        from .customization import ExtCustomization
        if self._client.match_app_version(label='widget', version='>=3.0.0'):
            raise DeprecationWarning("Customizations are deprecated. We introduced the `Widget` concept in version 3.")

        # For now, we only allow customization in an Ext JS context
        return ExtCustomization(activity=self, client=self._client)

    def configure(self, inputs, outputs):
        """
        Configure activity input and output.

        You need to provide a list of input and output :class:`Property`. Does not work with lists of propery id's.

        :param inputs: iterable of input property models
        :type inputs: list(:class:`Property`)
        :param outputs: iterable of output property models
        :type outputs: list(:class:`Property`)
        :raises APIError: when unable to configure the activity
        """
        def _get_propertyset(proplist):
            """Make it into a unique list of properties to configure for either inputs or outputs."""
            from pykechain.models import Property
            propertyset = []
            for property in proplist:
                if isinstance(property, Property):
                    propertyset.append(property.id)
                elif is_uuid(property):
                    propertyset.append(property)
            return list(set(propertyset))

        url = self._client._build_url('activity', activity_id='{}/update_associations'.format(self.id))

        if not all([p._json_data.get('category') == Category.MODEL for p in inputs]) and \
                not all([p._json_data.get('category') == Category.MODEL for p in outputs]):
            raise IllegalArgumentError('All Properties need to be of category MODEL to configure a task')

        response = self._client._request('PUT', url, json={
            'inputs': _get_propertyset(inputs),
            'outputs': _get_propertyset(outputs)
        })

        if response.status_code != requests.codes.ok:  # pragma: no cover
            raise APIError("Could not configure activity")

        self.refresh(json=response.json().get('results')[0])

    def download_as_pdf(self, target_dir=None, pdf_filename=None, paper_size=PaperSize.A4,
                        paper_orientation=PaperOrientation.PORTRAIT, include_appendices=False):
        """
        Retrieve the PDF of the Activity.

        .. versionadded:: 2.1

        :param target_dir: (optional) directory path name where the store the log.txt to.
        :type target_dir: basestring or None
        :param pdf_filename: (optional) log filename to write the log to, defaults to `log.txt`.
        :type pdf_filename: basestring or None
        :param paper_size: The size of the paper to which the PDF is downloaded:
                               - a4paper (default): A4 paper size
                               - a3paper: A3 paper size
                               - a2paper: A2 paper size
                               - a1paper: A1 paper size
                               - a0paper: A0 paper size
        :type paper_size: basestring (see :class:`enums.PaperSize`)
        :param paper_orientation: The orientation of the paper to which the PDF is downloaded:
                               - portrait (default): portrait orientation
                               - landscape: landscape orientation
        :type paper_size: basestring (see :class:`enums.PaperOrientation`)
        :param include_appendices: True if the PDF should contain appendices, False (default) if otherwise.
        :type include_appendices: bool
        :raises APIError: if the pdf file could not be found.
        :raises OSError: if the file could not be written.
        """
        if not pdf_filename:
            pdf_filename = self.name + '.pdf'
        if not pdf_filename.endswith('.pdf'):
            pdf_filename += '.pdf'

        full_path = os.path.join(target_dir or os.getcwd(), pdf_filename)

        request_params = {
            'papersize': paper_size,
            'orientation': paper_orientation,
            'appendices': include_appendices
        }

        url = self._client._build_url('activity_export', activity_id=self.id)
        response = self._client._request('GET', url, params=request_params)
        if response.status_code != requests.codes.ok:  # pragma: no cover
            raise APIError("Could not download PDF of activity '{}': '{}'".format(self.name, response.content))

        # If appendices are included, the request becomes asynchronous

        if include_appendices:
            data = response.json()

            # Download the pdf async
            url = urljoin(self._client.api_root, data['download_url'])

            count = 0

            while count <= ASYNC_TIMEOUT_LIMIT:
                response = self._client._request('GET', url=url)

                if response.status_code == requests.codes.ok:  # pragma: no cover
                    with open(full_path, 'wb') as f:
                        for chunk in response.iter_content(1024):
                            f.write(chunk)
                    return

                count += ASYNC_REFRESH_INTERVAL
                time.sleep(ASYNC_REFRESH_INTERVAL)

            raise APIError("Could not download PDF of activity {} within the time-out limit of {} "
                           "seconds".format(self.name, ASYNC_TIMEOUT_LIMIT))

        with open(full_path, 'wb') as f:
            for chunk in response.iter_content(1024):
                f.write(chunk)

    def move(self, parent, classification=None):
        """
        Move the `Activity` to a new parent.

        See :func:`pykechain.Client.move_activity` for available parameters.

        If you want to move an Activity from one classification to another, you need to provide the target
        classification. The classificaiton of the parent should match the one provided in the function. This is
        to ensure that you really want this to happen.

        :param parent: parent object to move activity to
        :type parent: Union[Activity2, Text]
        :param classifiction: (optional) classification of the target parent if you want to change the classification.
        :type classification: ActivityClassification or None
        :raises IllegalArgumentError: if the 'parent' activity_type is not :class:`enums.ActivityType.SUBPROCESS`
        :raises IllegalArgumentError: if the 'parent' type is not :class:`Activity2` or UUID
        :raises APIError: if an Error occurs.
        """
        return self._client.move_activity(self, parent, classification=classification)
